"""
Setup Library Class
----------------


**Description**


This class contains all the functionalities to create and edit a setup. It is based on templates in order to allow to create and modify setup properties easily


================

"""
from __future__ import absolute_import

from collections import OrderedDict
import os.path

from ..generic.general_methods import aedt_exception_handler, generate_unique_name

from .SetupTemplates import SweepHFSS, SweepQ3D, SetupKeys, SweepHFSS3DLayout
from ..application.DataHandlers import tuple2dict, dict2arg


class Setup(object):
    """This Class defines the functions to initialize, create and update a Setup in Electronic Desktop"""

    @property
    def parent(self):
        """ """
        return self._parent

    @parent.setter
    def parent(self, value):
        """

        Parameters
        ----------
        value :
            

        Returns
        -------

        """
        self._parent = value
        pass

    @property
    def omodule(self):
        """ """
        return self._parent.oanalysis

    def __repr__(self):
        return "SetupName " + self.name + " with " + str(len(self.sweeps)) + " Sweeps"

    def __init__(self, parent, solutiontype, setupname="MySetupAuto", isnewsetup=True):
        """
        :param parent: AEDT Module for Analysis Setup
        :param setupname: Setup Name
        :param solutiontype: Setup Type of Application.SolutionType
        :param isnewsetup: Boolean. True if is a new setup to be created from template. False to access existing Setup
        """
        self._parent = None
        self.parent = parent
        self.setuptype = solutiontype
        self.name = setupname
        self.props = {}
        self.sweeps = []
        if isnewsetup:
            setup_template = SetupKeys.SetupTemplates[solutiontype]
            for t in setup_template:
                tuple2dict(t, self.props)
        else:
            try:
                setups_data = self.parent.design_properties["AnalysisSetup"]["SolveSetups"]
                if setupname in setups_data:
                    setup_data = setups_data[setupname]
                    if "Sweeps" in setup_data and self.setuptype not in [0,7]:   #0 and 7 represent setup HFSSDrivenAuto
                        if self.setuptype <= 4:
                            app = setup_data["Sweeps"]
                            app.pop('NextUniqueID', None)
                            app.pop('MoveBackForward', None)
                            app.pop('MoveBackwards', None)
                            for el in app:
                                if type(app[el]) is OrderedDict:
                                    self.sweeps.append(SweepHFSS(self.omodule, setupname, el, props=app[el]))

                        else:
                            app = setup_data["Sweeps"]
                            for el in app:
                                if type(app[el]) is OrderedDict:
                                    self.sweeps.append(SweepQ3D(self.omodule, setupname, el, props=app[el]))
                        setup_data.pop('Sweeps', None)
                    self.props = OrderedDict(setup_data)
            except:
                self.props = OrderedDict()

    @aedt_exception_handler
    def create(self):
        """Insert a new Setup based on Class settings into the AEDT Application
        
        :return: the argument list

        Parameters
        ----------

        Returns
        -------

        """
        soltype = SetupKeys.SetupNames[self.setuptype]
        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)
        self.omodule.InsertSetup(soltype, arg)
        return arg

    @aedt_exception_handler
    def update(self, update_dictionary=None):
        """update the setup based on the class argument or to the updated dictionary passed as arcument

        Parameters
        ----------
        update_dictionary :
            optional dictionary of settings to apply (Default value = None)

        Returns
        -------
        type
            bool

        """
        if update_dictionary:
            for el in update_dictionary:
                self.props[el]=update_dictionary[el]
        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)

        self.omodule.EditSetup(self.name, arg)
        return True

    @aedt_exception_handler
    def _expression_cache(self, expression_list, report_type_list,intrinsics_list, isconvergence_list,
                          isrelativeconvergence, conv_criteria):
        """

        Parameters
        ----------
        expression_list :
            
        report_type_list :
            
        intrinsics_list :
            
        isconvergence_list :
            
        isrelativeconvergence :
            
        conv_criteria :
            

        Returns
        -------

        """
        if isrelativeconvergence:
            userelative = 1
        else:
            userelative = 0

        list_data = ["NAME:ExpressionCache"]
        if type(expression_list) is list:
            i=0
            while i < len(expression_list):
                expression=expression_list[i]
                name = expression.replace("(", "_") + "1"
                name = name.replace(")", "_")
                name = name.replace(" ", "_")
                if type(report_type_list) is list:
                    report_type = report_type_list[i]
                else:
                    report_type = report_type_list
                if type(isconvergence_list) is list:
                    isconvergence = isconvergence_list[i]
                else:
                    isconvergence = isconvergence_list
                if type(intrinsics_list) is list:
                    intrinsics = intrinsics_list[i]
                else:
                    intrinsics = intrinsics_list
                list_data.append([
                    "NAME:CacheItem",
                    "Title:=", name,
                    "Expression:=", expression,
                    "Intrinsics:=", intrinsics,
                    "IsConvergence:=", isconvergence,
                    "UseRelativeConvergence:=", 1,
                    "MaxConvergenceDelta:=", 1,
                    "MaxConvergeValue:=", "0.01",
                    "ReportType:=", report_type,
                    [
                        "NAME:ExpressionContext"
                    ]
                ])
                i+=1

        else:
            name = expression_list.replace("(", "") + "1"
            name = name.replace(")", "")
            name = name.replace(" ", "")
            name = name.replace(",", "_")
            list_data.append([
                "NAME:CacheItem",
                "Title:=", name,
                "Expression:=", expression_list,
                "Intrinsics:=", intrinsics_list,
                "IsConvergence:=", isconvergence_list,
                "UseRelativeConvergence:=", userelative,
                "MaxConvergenceDelta:=", conv_criteria,
                "MaxConvergeValue:=", str(conv_criteria),
                "ReportType:=", report_type_list,
                [
                    "NAME:ExpressionContext"
                ]
            ])

        return list_data

    @aedt_exception_handler
    def enable_expression_cache(self, expressions, report_type="Fields", intrinsics='', isconvergence=True,
                                isrelativeconvergence=True, conv_criteria=1):
        """Enable Setup Expression Cache

        Parameters
        ----------
        conv_criteria :
            param isrelativeconvergence: (Default value = 1)
        expressions :
            Formula to be added to the expression cache. It can be a list or a string
        report_type :
            Report type of expression cache (eg. Fields). It can be a list or a string (same length of expression_list) (Default value = "Fields")
        isconvergence :
            Boolean if expression is in convergence criteria. It can be a list or a string (same length of expression_list) (Default value = True)
        intrinsics :
            List of Intrinsics of expression (if any). It can be a list or a string (same length of expression_list) (Default value = '')
        isrelativeconvergence :
             (Default value = True)

        Returns
        -------
        type
            none

        """
        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)
        expression_cache = self._expression_cache(expressions, report_type, intrinsics, isconvergence,
                                                  isrelativeconvergence, conv_criteria)
        arg.append(expression_cache)
        self.omodule.EditSetup(self.name, arg)
        return True

    @aedt_exception_handler
    def add_derivatives(self, derivative_list):
        """Add Derivatives to setup

        Parameters
        ----------
        derivative_list :
            derivative list

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)
        arg.append("VariablesForDerivatives:=")
        arg.append(derivative_list)
        self.omodule.EditSetup(self.name, arg)
        return True

    @aedt_exception_handler
    def enable(self, setup_name=None):
        """Enable a specific setup.

        Parameters
        ----------
        setup_name : str, optional
            Name of the setup, by default ``None``

        Returns
        -------
        bool
            ``True`` if successful, ``False`` otherwise.
        """
        if not setup_name:
            setup_name = self.name

        self.omodule.EditSetup(
            setup_name,
            [
                "NAME:" + setup_name,
                "IsEnabled:=", True
            ]
        )
        return True

    @aedt_exception_handler
    def disable(self, setup_name=None):
        """Disable a specific setup.

        Parameters
        ----------
        setup_name : str, optional
            Name of the setup, by default ``None``

        Returns
        -------
        bool
            ``True`` if successful, ``False`` otherwise.
        """
        if not setup_name:
            setup_name = self.name

        self.omodule.EditSetup(
            setup_name,
            [
                "NAME:" + setup_name,
                "IsEnabled:", False
            ]
        )
        return True

    @aedt_exception_handler
    def add_sweep(self, sweepname=None, sweeptype= "Interpolating"):
        """Add a Sweep to the Project

        Parameters
        ----------
        sweepname :
            str Sweep Name (Default value = None)
        sweeptype :
            sweep type (Default value = "Interpolating")

        Returns
        -------
        type
            sweep object

        """
        if not sweepname:
            sweepname = generate_unique_name("Sweep")
        if self.setuptype <= 4:
            sweep_n = SweepHFSS(self.omodule, self.name, sweepname, sweeptype)
        else:
            sweep_n = SweepQ3D(self.omodule, self.name, sweepname, sweeptype)
        sweep_n.create()
        self.sweeps.append(sweep_n)
        return sweep_n

    @aedt_exception_handler
    def add_mesh_link(self, design_name, solution_name, parameters_dict, project_name="This Project*"):
        """Add Mesh Link to another design. Design can be in the same project (default) or external project

        Parameters
        ----------
        project_name :
            name of project "This Project*" by default
        design_name :
            str name of the design
        solution_name :
            str, name of solution in the format "setupname : solutionname". Optionally use appname.nominal_adaptive to get nominal adaptive or appname.nominal_sweep
        parameters_dict :
            dictionary of parameters. Optionally use appname.available_variations.nominal_w_values_dict property to get nominal values

        Returns
        -------
        type
            Bool

        """
        meshlinks = self.props["MeshLink"]
        meshlinks["ImportMesh"] = True
        meshlinks["Project"] = project_name
        meshlinks["Product"] = "ElectronicsDesktop"
        meshlinks["Design"] = design_name
        meshlinks["Soln"] = solution_name
        meshlinks["Params"] = OrderedDict({})
        for el in parameters_dict:
            if el in list(self._parent.available_variations.nominal_w_values_dict.keys()):
                meshlinks["Params"][el] = el
            else:
                meshlinks["Params"][el] = parameters_dict[el]
        meshlinks["ForceSourceToSolve"] = True
        meshlinks["PreservePartnerSoln"] = True
        meshlinks["PathRelativeTo"] = "TargetProject"
        meshlinks["ApplyMeshOp"] = True
        self.update()
        return True

class SetupCircuit(object):
    """This Class defines the functions to initialize, create and update a Setup in Electronic Desktop

    Parameters
    ----------
    design :
        AEDT oDesign module
    setupmodule :
        AEDT Module for Analysis Setup
    setupname :
        Setup Name
    solutiontype :
        Setup Type of Application.SolutionType
    isnewsetup :
        Boolean. True if is a new setup to be created from template. False to access existing Setup

    Returns
    -------

    """
    @property
    def name(self):
        """ """
        return self._Name

    @name.setter
    def name(self, name):
        """

        Parameters
        ----------
        name :
            

        Returns
        -------

        """
        self._Name = name
        self.props["Name"] = name

    @property
    def parent(self):
        """ """
        return self._parent

    @parent.setter
    def parent(self, name):
        """

        Parameters
        ----------
        name :
            

        Returns
        -------

        """
        self._parent = name

    @property
    def odesign(self):
        """ """
        return self._parent.odesign

    @property
    def omodule(self):
        """ """
        return self._parent.oanalysis

    def __init__(self, parent,  solutiontype, setupname="MySetupAuto", isnewsetup=True):
        self._parent = None
        self.parent = parent
        self.setuptype = solutiontype
        self._Name = "LinearFrequency"
        self.props = {}
        if isnewsetup:
            setup_template = SetupKeys.SetupTemplates[solutiontype]
            for t in setup_template:
                tuple2dict(t, self.props)
        else:
            try:
                setups_data = self.parent.design_properties["SimSetups"]["SimSetup"]
                if type(setups_data) is not list:
                    setups_data=[setups_data]
                for setup in setups_data:
                    if setupname == setup["Name"]:
                        setup_data = setup
                        setup_data.pop('Sweeps', None)
                        self.props = setup_data
            except:
                self.props = {}
        self.name = setupname

    @aedt_exception_handler
    def create(self):
        """Insert a new Setup based on Class settings into the AEDT Application
        
        :return: the argument list

        Parameters
        ----------

        Returns
        -------

        """
        soltype = SetupKeys.SetupNames[self.setuptype]
        arg = ["NAME:SimSetup"]
        dict2arg(self.props, arg)
        self._setup(soltype, arg)
        return arg

    @aedt_exception_handler
    def _setup(self, soltype, arg, newsetup=True):
        """

        Parameters
        ----------
        soltype :
            
        arg :
            
        newsetup :
             (Default value = True)

        Returns
        -------

        """
        if newsetup:
            if soltype == "NexximLNA":
                self.omodule.AddLinearNetworkAnalysis(arg)
            elif soltype == "NexximDC":
                self.omodule.AddDCAnalysis(arg)
            elif soltype == "NexximTransient":
                self.omodule.AddTransient(arg)
            else:
                print("Not Implemented Yet")
        else:
            if soltype == "NexximLNA":
                self.omodule.EditLinearNetworkAnalysis(self.name, arg)
            elif soltype == "NexximDC":
                self.omodule.EditDCAnalysis(self.name, arg)
            elif soltype == "NexximTransient":
                self.omodule.EditTransient(self.name, arg)
            else:
                print("Not Implemented Yet")
        return True

    @aedt_exception_handler
    def update(self, update_dictionary=None):
        """update the setup based on the class argument or to the updated dictionary passed as arcument

        Parameters
        ----------
        update_dictionary :
            optional dictionary of settings to apply (Default value = None)

        Returns
        -------
        type
            bool

        """
        if update_dictionary:
            for el in update_dictionary:
                self.props[el]=update_dictionary[el]
        arg = ["NAME:SimSetup"]
        soltype = SetupKeys.SetupNames[self.setuptype]
        dict2arg(self.props, arg)
        self._setup(soltype, arg, False)
        return True

    @aedt_exception_handler
    def _expression_cache(self, expression_list, report_type_list, intrinsics_list, isconvergence_list,
                          isrelativeconvergence, conv_criteria):
        """

        Parameters
        ----------
        expression_list :
            
        report_type_list :
            
        intrinsics_list :
            
        isconvergence_list :
            
        isrelativeconvergence :
            
        conv_criteria :
            

        Returns
        -------

        """

        if isrelativeconvergence:
            userelative = 1
        else:
            userelative = 0

        list_data = ["NAME:ExpressionCache"]
        if type(expression_list) is list:
            i=0
            while i < len(expression_list):
                expression=expression_list[i]
                name = expression.replace("(", "_") + "1"
                name = name.replace(")", "_")
                name = name.replace(" ", "_")
                if type(report_type_list) is list:
                    report_type = report_type_list[i]
                else:
                    report_type = report_type_list
                if type(isconvergence_list) is list:
                    isconvergence = isconvergence_list[i]
                else:
                    isconvergence = isconvergence_list
                if type(intrinsics_list) is list:
                    intrinsics = intrinsics_list[i]
                else:
                    intrinsics = intrinsics_list
                list_data.append([
                    "NAME:CacheItem",
                    "Title:=", name,
                    "Expression:=", expression,
                    "Intrinsics:=", intrinsics,
                    "IsConvergence:=", isconvergence,
                    "UseRelativeConvergence:=", 1,
                    "MaxConvergenceDelta:=", 1,
                    "MaxConvergeValue:=", "0.01",
                    "ReportType:=", report_type,
                    [
                        "NAME:ExpressionContext"
                    ]
                ])
                i+=1

        else:
            name = expression_list.replace("(", "") + "1"
            name = name.replace(")", "")
            name = name.replace(" ", "")
            name = name.replace(",", "_")
            list_data.append([
                "NAME:CacheItem",
                "Title:=", name,
                "Expression:=", expression_list,
                "Intrinsics:=", intrinsics_list,
                "IsConvergence:=", isconvergence_list,
                "UseRelativeConvergence:=", userelative,
                "MaxConvergenceDelta:=", conv_criteria,
                "MaxConvergeValue:=", str(conv_criteria),
                "ReportType:=", report_type_list,
                [
                    "NAME:ExpressionContext"
                ]
            ])

        return list_data

    @aedt_exception_handler
    def enable_expression_cache(self, expressions, report_type="Fields", intrinsics='', isconvergence=True,
                                isrelativeconvergence=True, conv_criteria=1):
        """

        Parameters
        ----------
        conv_criteria :
            param isrelativeconvergence: (Default value = 1)
        expressions :
            Formula to be added to the expression cache. It can be a list or a string
        report_type :
            Report type of expression cache (eg. Fields). It can be a list or a string (same length of expression_list) (Default value = "Fields")
        isconvergence :
            Boolean if expression is in convergence criteria. It can be a list or a string (same length of expression_list) (Default value = True)
        intrinsics :
            List of Intrinsics of expression (if any). It can be a list or a string (same length of expression_list) (Default value = '')
        isrelativeconvergence :
             (Default value = True)

        Returns
        -------
        type
            none

        """
        arg = ["Name:SimSetup"]
        dict2arg(self.props, arg)
        expression_cache = self._expression_cache(expressions, report_type, intrinsics, isconvergence,
                                                  isrelativeconvergence, conv_criteria)
        arg.append(expression_cache)
        self.omodule.EditSetup(self.name, arg)
        return True

    @aedt_exception_handler
    def add_derivatives(self, derivative_list):
        """Add Derivatives to Setup

        Parameters
        ----------
        derivative_list :
            derivative lists

        Returns
        -------
        type
            Bool

        """
        arg = ["Name:SimSetup"]
        dict2arg(self.props, arg)
        arg.append("VariablesForDerivatives:=")
        arg.append(derivative_list)
        self.omodule.EditSetup(self.name, arg)
        return True

    @aedt_exception_handler
    def enable(self, setup_name=None):
        """Enable a specific setup.

        Parameters
        ----------
        setup_name : str, optional
            Name of the setup, by default ``None``

        Returns
        -------
        bool
            ``True`` if successful, ``False`` otherwise.
        """
        if not setup_name:
            setup_name = self.name
        self.odesign.EnableSolutionSetup(setup_name, True)
        return True

    @aedt_exception_handler
    def disable(self, setup_name=None):
        """Disable a specific setup.

        Parameters
        ----------
        setup_name : str, optional
            Name of the setup, by default ``None``

        Returns
        -------
        bool
            ``True`` if successful, ``False`` otherwise.
        """
        if not setup_name:
            setup_name = self.name
        self.odesign.EnableSolutionSetup(setup_name, False)
        return True


class Setup3DLayout(object):
    """This Class defines the functions to initialize, create and update a Setup in Electronic Desktop

    Parameters
    ----------
    setupmodule :
        AEDT Module for Analysis Setup
    setupname :
        Setup Name
    solutiontype :
        Setup Type of Application.SolutionType
    isnewsetup :
        Boolean. True if is a new setup to be created from template. False to access existing Setup

    Returns
    -------

    """

    @property
    def omodule(self):
        """ """
        return self.parent.oanalysis

    def __init__(self, parent, solutiontype, setupname="MySetupAuto", isnewsetup=True):
        self.parent = parent
        self._solutiontype = solutiontype
        self.name = setupname
        self.props = OrderedDict()
        self.sweeps = []
        if isnewsetup:
            setup_template = SetupKeys.SetupTemplates[self._solutiontype]
            for t in setup_template:
                tuple2dict(t, self.props)
        else:
            try:
                setups_data = self.parent.design_properties["Setup"]["Data"]
                if setupname in setups_data:
                    setup_data = setups_data[setupname]
                    if "Data" in setup_data:   #0 and 7 represent setup HFSSDrivenAuto
                        app = setup_data["Data"]
                        for el in app:
                            if type(app[el]) is OrderedDict:
                                self.sweeps.append(SweepHFSS3DLayout(self.omodule, setupname, el, props=app[el]))

                    self.props = OrderedDict(setup_data)
            except:
                self.props = OrderedDict()
            #
            # setup_data = self.omodule.GetSetupData(setupname)
            # dict_data = OrderedDict()
            # _arg2dict(setup_data, dict_data)
            # self.props = dict_data[setupname]

    @property
    def setup_type(self):
        """:return: setup type"""
        if 'SolveSetupType' in self.props:
            return self.props['SolveSetupType']
        else:
            return None

    @aedt_exception_handler
    def create(self):
        """Insert a new Setup based on Class settings into the AEDT Application
        
        :return: the argument list

        Parameters
        ----------

        Returns
        -------

        """
        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)
        self.omodule.Add(arg)
        return True

    @aedt_exception_handler
    def update(self):
        """update the setup based on the class argument or to the updated dictionary passed as arcument

        Parameters
        ----------
        update_dictionary :
            optional dictionary of settings to apply

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)
        self.omodule.Edit(self.name, arg)
        return True

    @aedt_exception_handler
    def enable(self):
        """Enable specific Setup

        Parameters
        ----------
        setup_name :
            optional setup name

        Returns
        -------
        type
            Bool

        """
        self.props['Properties']['Enable'] = "true"
        self.update()
        return True

    @aedt_exception_handler
    def disable(self):
        """Disable specific Setup

        Parameters
        ----------
        setup_name :
            optional setup name

        Returns
        -------
        type
            Bool

        """
        self.props['Properties']['Enable'] = "false"
        self.update()
        return True

    @aedt_exception_handler
    def export_to_hfss(self, file_fullname):
        """Export Project to

        Parameters
        ----------
        file_fullname :
            fullname of desgintation

        Returns
        -------
        type
            Bool

        """

        file_fullname = os.path.normpath(file_fullname)
        if not os.path.isdir(os.path.dirname(file_fullname)):
            return False
        file_fullname = os.path.splitext(file_fullname)[0] + '.aedt'
        self.omodule.ExportToHfss(self.name, file_fullname)
        return True


    @aedt_exception_handler
    def add_sweep(self, sweepname=None, sweeptype="Interpolating"):
        """Add Frequency Sweep

        Parameters
        ----------
        sweepname :
            str sweep name (Default value = None)
        sweeptype :
            str sweep type (Default value = "Interpolating")

        Returns
        -------
        SweeHFSS3DLayout
            sweep object

        """
        if not sweepname:
            sweepname = generate_unique_name("Sweep")
        sweep_n = SweepHFSS3DLayout(self.omodule, self.name, sweepname, sweeptype)
        if sweep_n.create():
            self.sweeps.append(sweep_n)
            return sweep_n
        return False



import math

from ..generic.general_methods import aedt_exception_handler
from .Modeler import Modeler, GeometryModeler
from .Primitives2D import Primitives2D


class ModelerRMxprt(Modeler):
    """ """

    def __init__(self, parent):
        Modeler.__init__(self, parent)

    @property
    def oeditor(self):
        """ """
        return self.odesign.SetActiveEditor("Machine")


class Modeler2D(GeometryModeler):
    """ """

    def __init__(self, application):
        GeometryModeler.__init__(self, application, is3d=False)
        self._primitives = Primitives2D(self._parent, self)
        self._primitivesDes = self._parent.project_name + self._parent.design_name

    def __get__(self, instance, owner):
        self._parent = instance
        return self

    @property
    def primitives(self):
        """ """
        if self._primitivesDes != self._parent.project_name + self._parent.design_name:
            self._primitives = Primitives2D(self._parent, self)
            self._primitivesDes = self._parent.project_name + self._parent.design_name
        return self._primitives

    @aedt_exception_handler
    def calculate_radius_2D(self, object_name, inner=False):
        """Calculate the extremity of an object in the radial direction. 

        Parameters
        ----------
        object_name : str
            name of the object from which to calculate the radius.
        inner : bool, optional
             The default is ``False``.

        Returns
        -------
        type
            Radius.
            .. note::
                If ``inner=True``, then the maximum is returned; otherwise,
                the minimum is returned.

        """
        oVertexIDs = self.oeditor.GetVertexIDsFromObject(object_name)
        vert_id_int = [int(x) for x in oVertexIDs]
        if inner:
            radius = 0
        else:
            radius = 1e9

        for vertex in vert_id_int:
            pos = self.oeditor.GetVertexPosition(vertex)
            vertex_radius = math.sqrt(float(pos[0]) ** 2 + float(pos[1]) ** 2)
            if inner:
                if vertex_radius > radius:
                    radius = vertex_radius
            else:
                if vertex_radius < radius:
                    radius = vertex_radius

        return radius

    @aedt_exception_handler
    def radial_split_2D(self, radius, name):
        """Split the stator and rotor for mesh refinement.

        Parameters
        ----------
        radius :
            
        name :
            

        Returns
        -------

        """
        self.oeditor.CreateCircle(
            [
                "NAME:CircleParameters",
                "IsCovered:=", True,
                "XCenter:=", "0mm",
                "YCenter:=", "0mm",
                "ZCenter:=", "0mm",
                "Radius:=", radius,
                "WhichAxis:=", "Z",
                "NumSegments:=", "0"
            ],
            [
                "NAME:Attributes",
                "Name:=", name + "_split",
                "Flags:=", "",
                "Color:=", "(132 132 193)",
                "Transparency:=", 0,
                "PartCoordinateSystem:=", "Global",
                "UDMId:=", "",
                "Materiaobjidue:=", "\"vacuum\"",
                "SolveInside:=", True
            ])

        self.oeditor.Copy(
            [
                "NAME:Selections",
                "Selections:=", name
            ])

        self.oeditor.Paste()
        self.oeditor.Intersect(
            [
                "NAME:Selections",
                "Selections:=", "{0}1,{0}_split".format(name)
            ],
            [
                "NAME:IntersectParameters",
                "KeepOriginals:=", False
            ])

        self.subtract(name, name + "1")
        return True



from ..generic.general_methods import aedt_exception_handler, generate_unique_name
from .Analysis import Analysis
from .Design import solutions_settings, Design
from ..modeler.Circuit import ModelerEmit
from ..modules.PostProcessor import PostProcessor
from ..modules.SetupTemplates import SetupKeys
from ..modules.SolveSetup import SetupCircuit


class FieldAnalysisEmit(Design):
    """**AEDT Emit Analysis**

    Class for Emit Analysis Setup (Emit)

    It is automatically initialized by Application call (like HFSS,
    Q3D...). Refer to Application function for inputs definition

    Parameters
    ----------

    Returns
    -------

    """
    @property
    def solution_type(self):
        """ """
        return self._solution_type


    @solution_type.setter
    def solution_type(self, soltype):
        """Solution Type

        Parameters
        ----------
        soltype :
            SolutionType object

        Returns
        -------

        """
        if soltype:
            self._solution_type = "EMIT"
        else:
            self._solution_type = "EMIT"

    @property
    def existing_analysis_setups(self):
        """ """
        return []

    @property
    def setup_names(self):
        """ """
        return []


    def __init__(self, application, projectname, designname, solution_type, setup_name=None,
                 specified_version=None, NG=False, AlwaysNew=True, release_on_exit=True):
        self.solution_type = solution_type
        Design.__init__(self, application, projectname, designname, solution_type,
                        specified_version, NG, AlwaysNew, release_on_exit)
        self._modeler = ModelerEmit(self)
        self._post = PostProcessor(self)

    @property
    def modeler(self):
        """:return: Design oModeler"""
        return self._modeler

    @property
    def oanalysis(self):
        """:return: Design Module "SimSetup"
        """
        return None




"""
Design Module
-------------

Description
===========
This class contains all the basic Project information and
objects. since it is hinerithed in the Main Tool class it will be a
simple call from it

Examples
--------

>>> from pyaedt import Hfss
>>> hfss = Hfss()

Return the oProject object

>>> hfss.oproject


"""
from __future__ import absolute_import

import warnings
import os
import re
import csv
import shutil
import sys
import json
import string
import random
import time
import logging
from collections import OrderedDict
from .MessageManager import AEDTMessageManager
from .Variables import VariableManager, DataSet
from ..desktop import exception_to_desktop, Desktop, force_close_desktop, release_desktop, get_version_env_variable
from ..generic.LoadAEDTFile import load_entire_aedt_file
from ..generic.general_methods import aedt_exception_handler
from ..generic.list_handling import variation_string_to_dict
from ..modules.Boundary import BoundaryObject
from ..generic.general_methods import generate_unique_name


design_solutions = {
    "Maxwell 2D": [
        "MagnetostaticXY",
        "MagnetostaticZ",
        "EddyCurrentXY",
        "EddyCurrentZ",
        "TransientXY",
        "TransientZ",
        "ElectrostaticXY",
        "ElectrostaticZ",
        "ElectricTransientXY",
        "ElectricTransientZ",
        "ElectroDCConductionXY",
        "ElectroDCConductionZ"
    ],
    "Maxwell 3D": [
        "Magnetostatic",
        "EddyCurrent",
        "Transient",
        "Electrostatic",
        "DCConduction",
        "ElectroDCConduction",
        "ElectricTransient"
    ],
    "Twin Builder": [
        "TR",
        "AC",
        "DC"
    ],
    "Circuit Design": [
        "NexximLNA"
    ],
    "2D Extractor": [
        "Open",
        "Closed"
    ],
    "Q3D Extractor": [
        "Q3D Extractor"
    ],
    "HFSS": [
        "DrivenModal",
        "DrivenTerminal",
        "Transient Network",
        "Eigenmode",
        "Characteristic Mode",
        "SBR+"
    ],
    "Icepak": [
        "SteadyTemperatureAndFlow",
        "SteadyTemperatureOnly",
        "SteadyFlowOnly",
        "TransientTemperatureAndFlow",
        "TransientTemperatureOnly",
        "TransientFlowOnly"

    ],
    "RMxprtSolution": [
        "IRIM", "ORIM", "SRIM", "WRIM", "DFIG", "AFIM", "HM", "RFSM", "RASM", "RSM", "ISM", "APSM", "IBDM", "ABDM",
        "TPIM", "SPIM", "TPSM", "BLDC", "ASSM", "PMDC", "SRM", "LSSM", "UNIM", "DCM", "CPSM", "NSSM"
    ],
    "ModelCreation": [
        "IRIM", "ORIM", "SRIM", "WRIM", "DFIG", "AFIM", "HM", "RFSM", "RASM", "RSM", "ISM", "APSM", "IBDM", "ABDM",
    ],
    "HFSS 3D Layout Design": [
        ""
    ],
    "Mechanical": [
        "Thermal",
        "Modal",
        "Structural"
    ],
    "EMIT": ["EMIT"]
}

solutions_settings = {
    "DrivenModal": "DrivenModal",
    "DrivenTerminal": "DrivenTerminal",
    "EigenMode": "EigenMode",
    "Transient Network": "Transient Network",
    "SBR+": "SBR+",
    "Transient": "Transient",
    "Magnetostatic": "Magnetostatic",
    "EddyCurrent": "EddyCurrent",
    "Electrostatic": "Electrostatic",
    "ElectroDCConduction": "ElectroDCConduction",
    "ElectricTransient": "ElectricTransient",
    "Matrix": "Matrix",
    "SteadyTemperatureAndFlow": [
        "NAME:SolutionTypeOption",
        "SolutionTypeOption:=",
        "SteadyState",
        "ProblemOption:=",
        "SteadyTemperatureAndFlow"
    ],
    "SteadyTemperatureOnly": [
        "NAME:SolutionTypeOption",
        "SolutionTypeOption:=",
        "SteadyState",
        "ProblemOption:=",
        "SteadyTemperatureOnly"
    ],
    "SteadyFlowOnly": [
        "NAME:SolutionTypeOption",
        "SolutionTypeOption:=",
        "SteadyState",
        "ProblemOption:=",
        "SteadyFlowOnly"
    ],
    "TransientTemperatureAndFlow": [
        "NAME:SolutionTypeOption",
        "SolutionTypeOption:=",
        "Transient",
        "ProblemOption:=",
        "SteadyTemperatureAndFlow"
    ],
    "TransientTemperatureOnly": [
        "NAME:SolutionTypeOption",
        "SolutionTypeOption:=",
        "Transient",
        "ProblemOption:=",
        "SteadyTemperatureOnly"
    ],
    "TransientFlowOnly": [
        "NAME:SolutionTypeOption",
        "SolutionTypeOption:=",
        "Transient",
        "ProblemOption:=",
        "SteadyFlowOnly"
    ],
    "NexximLNA": "NexximLNA",
    "NexximDC": "NexximDC",
    "NexximTransient": "NexximTransient",
    "NexximQuickEye": "NexximQuickEye",
    "NexximVerifEye": "NexximVerifEye",
    "NexximAMI": "NexximAMI",
    "NexximOscillatorRSF": "NexximOscillatorRSF",
    "NexximOscillator1T": "NexximOscillator1T",
    "NexximOscillatorNT": "NexximOscillatorNT",
    "NexximHarmonicBalance1T": "NexximHarmonicBalance1T",
    "NexximHarmonicBalanceNT": "NexximHarmonicBalanceNT",
    "NexximSystem": "NexximSystem",
    "NexximTVNoise": "NexximTVNoise",
    "HSPICE": "HSPICE",
    "TR": "TR",
    "Open": "Open",
    "Closed": "Closed",
    "TransientXY": ["Transient", "XY"], "TransientZ": ["Transient", "about Z"],
    "MagnetostaticXY": ["Magnetostatic", "XY"], "MagnetostaticZ": ["Magnetostatic", "about Z"],
    "EddyCurrentXY": ["EddyCurrent", "XY"], "EddyCurrentZ": ["EddyCurrent", "about Z"],
    "ElectrostaticXY": ["Electrostatic", "XY"], "ElectrostaticZ": ["Electrostatic", "about Z"],
    "ElectroDCConductionXY": ["ElectroDCConduction", "XY"], "ElectroDCConductionZ": ["ElectroDCConduction", "about Z"],
    "ElectricTransientXY": ["ElectricTransient", "XY"], "ElectricTransientZ": ["ElectricTransient", "about Z"],
    "Modal": "Modal", "Thermal": "Thermal","Structural" : "Structural",
    "IRIM" : "IRIM", "ORIM": "ORIM", "SRIM": "SRIM", "WRIM" : "WRIM", "DFIG" : "DFIG", "AFIM": "AFIM", "HM" : "HM",
    "RFSM" :"RFSM", "RASM": "RASM", "RSM" : "RSM", "ISM" : "ISM", "APSM" : "APSM", "IBDM" : "IBDM", "ABDM" : "ABDM",
    "TPIM" : "TPIM", "SPIM" : "SPIM", "TPSM" : "TPSM", "BLDC" : "BLDC", "ASSM" : "ASSM", "PMDC" : "PMDC", "SRM" : "SRM",
    "LSSM" : "LSSM", "UNIM" : "UNIM", "DCM" : "DCM", "CPSM" : "CPSM", "NSSM" :"NSSM"
}

model_names = {
    "Maxwell 2D": "Maxwell2DModel",
    "Maxwell 3D": "Maxwell3DModel",
    "Twin Builder": "SimplorerCircuit",
    "Circuit Design": "NexximCircuit",
    "2D Extractor": "2DExtractorModel",
    "Q3D Extractor": "Q3DModel",
    "HFSS": "HFSSModel",
    "Mechanical": "MechanicalModel",
    "Icepak": "IcepakModel",
    "RMxprtSolution": "RMxprtDesign",
    "ModelCreation": "RMxprtDesign",
    "HFSS 3D Layout Design": "PlanarEMCircuit",
    "EMIT Design": "EMIT Design",
}


def list_difference(list1, list2):
    return list(set(list1) - set(list2))

class DesignCache(object):
    """Cache object to analyse the differences between the state of a design between two points in time.

    At present the contents of the design to be tracked are:

        * global-level messages in the message Manager
        * project-level messages in the message Manager
        * design-level messages in the message Manager
    """
    def __init__(self, parent):
        self._parent = parent
        self._allow_errors_local = []
        self._allow_errors_global = []
        self.clear()
        self._snapshot = self.design_snapshot()

    @property
    def allowed_error_messages(self):
        return self._allow_errors_global + self._allow_errors_local

    def ignore_error_message_local(self, msg):
        self._allow_errors_local.append("[error] {}".format(msg))

    def ignore_error_message_global(self, msg):
        self._allow_errors_global.append("[error] {}".format(msg))

    @property
    def no_change(self):
        """True if the design snapshot has not changed between updates"""
        return self._no_change

    @property
    def delta_global_messages(self):
        """A list of any new or missing global-level messages since the last update"""
        return self._delta_global_messages

    @property
    def delta_project_messages(self):
        """A list of any new or missing project-level messages since the last update"""
        return self._delta_global_messages

    @property
    def delta_design_messages(self):
        """A list of any new or missing design-level messages since the last update"""
        return self._delta_design_messages

    @property
    def delta_error_messages(self):
        """A list of any new or missing error messages since the last update"""
        return self._new_error_messages

    @property
    def no_new_messages(self):
        """Returns True if new messages have appeared since the last update or the message manager was cleared"""
        return not bool(self._delta_messages)

    @property
    def no_new_errors(self):
        """Returns True if new error messages have appeared since the last update"""
        return not bool(self._new_error_messages)

    @property
    def no_new_warnings(self):
        """Returns True if new error messages have appeared since the last update"""
        return not bool(self._new_warning_messages)

    @property
    def no_change(self):
        """Returns True if the cache elements have not changed since the last update"""
        return self.no_new_messages

    def design_snapshot(self):
        snapshot = {
            "Solids:":  self._parent.modeler.primitives.solids,
            "Lines:":  self._parent.modeler.primitives.lines,
            "Sheets": self._parent.modeler.primitives.sheets,
            "DesignName" : self._parent.design_name
        }
        return snapshot

    def clear(self):
        """Clear the cached values"""
        self._messages_global_level = []
        self._messages_project_level = []
        self._messages_design_level = []

    def update(self):
        """Get the current state values from the design and perform a delta-calculation with the cached values.
        Then replace the cached values with the current values. Note: this is done automatically when the property
        'no_change' is accessed!"""
        messages = self._parent._messenger.messages

        # Check whether the design snapshot has changed since the last update
        new_snapshot = self.design_snapshot()
        if new_snapshot == self._snapshot:
            self._no_change = True
        else:
            self._no_change = False

        self._snapshot = new_snapshot

        self._delta_global_messages = list_difference(messages.global_level, self._messages_global_level)
        self._delta_project_messages = list_difference(messages.project_level, self._messages_project_level)
        self._delta_design_messages = list_difference(messages.design_level, self._messages_design_level)
        self._delta_messages_unfiltered = self._delta_global_messages + self._delta_project_messages + self._delta_design_messages

        # filter out allowed messages
        self._delta_messages = []
        for msg in self._delta_messages_unfiltered:
            mask = False
            allowed_errors = self._allow_errors_local + self._allow_errors_global
            for allowed in allowed_errors:
                if msg.find(allowed) == 0:
                    mask = True
                    break
            if not mask:
                self._delta_messages.append(msg)

        self._new_error_messages = [msg for msg in self._delta_messages if msg.find("[error]") == 0]
        self._new_warning_messages = [msg for msg in self._delta_messages if msg.find("[warning]") == 0]

        self._messages_global_level = messages.global_level
        self._messages_project_level = messages.project_level
        self._messages_design_level = messages.design_level

        self._allow_errors_local = []

        return self


class Design(object):
    """Class Design. Contains all functions and objects connected to the active Project and Design"""
    def __str__(self):
        pyaedt_details = "      pyaedt API\n"
        pyaedt_details += "pyaedt running AEDT Version {} \n".format(self._aedt_version)
        pyaedt_details += "Running {} tool in AEDT\n".format(self.design_type)
        pyaedt_details += "Solution Type: {} \n".format(self.solution_type)
        pyaedt_details += "Project Name: {}    Design Name{} \n".format(self.project_name, self.design_name)
        pyaedt_details += "Project Path: \"{}\" \n".format(self.project_path)
        return pyaedt_details

    def __exit__(self, ex_type, ex_value, ex_traceback):
        if ex_type:
            exception_to_desktop(self, ex_value, ex_traceback)

    def __enter__(self):
        pass

    @aedt_exception_handler
    def __getitem__(self, variable_name):
        return self.variable_manager[variable_name].string_value

    @aedt_exception_handler
    def __setitem__(self, variable_name, variable_value):
        self.variable_manager[variable_name] = variable_value
        return True

    def __init__(self, design_type, project_name=None, design_name=None, solution_type=None,
                 specified_version=None, NG=False, AlwaysNew=False, release_on_exit=False):
        # Get Desktop from global Desktop Environment
        self._project_dictionary = OrderedDict()
        self.boundaries = OrderedDict()
        self.project_datasets = {}
        self.design_datasets = {}
        main_module = sys.modules['__main__']
        if "pyaedt_initialized" not in dir(main_module):
            Desktop(specified_version, NG, AlwaysNew, release_on_exit)
        self._project_dictionary = {}
        self._mttime = None
        self._desktop = main_module.oDesktop
        self._aedt_version = main_module.AEDTVersion
        self._desktop_install_dir = main_module.sDesktopinstallDirectory
        self._messenger = AEDTMessageManager(self)
        self.logger = logging.getLogger(__name__)

        assert design_type in design_solutions, "Invalid design type specified: {}".format(design_type)
        self._design_type = design_type
        if solution_type:
            assert solution_type in design_solutions[design_type], \
                "Invalid solution type {0} for design type {1}".format(solution_type, design_type)
        self._solution_type = solution_type
        self._odesign = None
        self._oproject = None
        self._design_type = design_type
        self.oproject = project_name
        self.odesign = design_name
        self._variable_manager = VariableManager(self)
        self.solution_type = self._solution_type
        self.project_datasets = self._get_project_datasets()
        self.design_datasets = self._get_design_datasets()

    @property
    def project_properies(self):
        """ """
        if os.path.exists(self.project_file):
            _mttime = os.path.getmtime(self.project_file)
            if _mttime != self._mttime:
                self._project_dictionary = load_entire_aedt_file(self.project_file)
                self._mttime = _mttime
        return self._project_dictionary

    @property
    def design_properties(self, design_name=None):
        """

        Parameters
        ----------
        design_name :
             (Default value = None)

        Returns
        -------

        """
        if not design_name:
            design_name = self.design_name
        try:
            if model_names[self._design_type] in self.project_properies['AnsoftProject']:
                designs = self.project_properies['AnsoftProject'][model_names[self._design_type]]
                if type(designs) is list:
                    for design in designs:
                        if design["Name"] == design_name:
                            return design
                else:
                    if designs["Name"] == design_name:
                        return designs
        except:
            return OrderedDict()


    @property
    def aedt_version_id(self):
        """Property

        :return: AEDT Version

        Parameters
        ----------

        Returns
        -------

        """
        version = self.odesktop.GetVersion()
        return get_version_env_variable(version)

    @property
    def design_name(self):
        """Name of the parent AEDT Design.

        Returns
        -------
        str
            Name of the parent AEDT Design.

        Examples
        --------
        Set the design name.

        >>> from pyaedt import Hfss
        >>> hfss = Hfss()
        >>> hfss.design_name = 'new_design'

        """
        if not self.odesign:
            return None
        name = self.odesign.GetName()
        if ";" in name:
            return name.split(";")[1]
        else:
            return name

    @design_name.setter
    def design_name(self, new_name):
        """
        Property

        :return: Change the name of the parent AEDT Design
        """

        if ";" in new_name:
            new_name = new_name.split(";")[1]

        self.odesign.RenameDesignInstance(self.design_name, new_name)
        timeout = 5.0
        timestep = 0.1
        while new_name not in [i.GetName() for i in list(self._oproject.GetDesigns())]:
            time.sleep(timestep)
            timeout -= timestep
            assert timeout >= 0

    @property
    def design_list(self):
        """List of available designs

        Returns
        -------

        """
        deslist = list(self.oproject.GetTopDesignList())
        updateddeslist = []
        for el in deslist:
            m = re.search(r'[^;]+$', el)
            updateddeslist.append(m.group(0))
        return updateddeslist

    @property
    def design_type(self):
        """Property

        :return: Design Type

        Parameters
        ----------

        Returns
        -------

        """
        #return self._odesign.GetDesignType()
        return self. _design_type


    @property
    def project_name(self):
        """Property

        :return: Project Name

        Parameters
        ----------

        Returns
        -------

        """
        if self._oproject:
            return self._oproject.GetName()
        else:
            return None

    @property
    def project_list(self):
        """Property

        :return: List of available Projects

        Parameters
        ----------

        Returns
        -------

        """
        return list(self._desktop.GetProjectList())

    @property
    def project_path(self):
        """Property

        :return: Project Path

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.normpath(self._oproject.GetPath())

    @property
    def project_file(self):
        """Property

        :return: Full absolute Project name and path

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.join(self.project_path, self.project_name + '.aedt')

    @property
    def lock_file(self):
        """Property

        :return: Full absolute Project lock file

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.join(self.project_path, self.project_name + '.aedt.lock')

    @property
    def results_directory(self):
        """Property

        :return: Full absolute path of the aedtresults directory

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.join(self.project_path, self.project_name + '.aedtresults')

    @property
    def solution_type(self):
        """Property

        :return: Solution Type

        Parameters
        ----------

        Returns
        -------

        """
        try:
            return self._odesign.GetSolutionType()
        except:
            if self.design_type == "Q3D Extractor":
                return "Matrix"
            elif self.design_type == "HFSS 3D Layout Design":
                return "HFSS3DLayout"
            else:
                return None

    @solution_type.setter
    def solution_type(self, soltype):
        """

        Parameters
        ----------
        soltype :
            SolutionType object

        Returns
        -------

        """

        if soltype:
            sol = solutions_settings[soltype]
            try:
                if self.design_type == "Maxwell 2D":
                    self.odesign.SetSolutionType(sol[0],sol[1])
                else:
                    self.odesign.SetSolutionType(sol)
            except:
                pass



    @property
    def valid_design(self):
        """Property

        :return: True if oproject and odesign exists

        Parameters
        ----------

        Returns
        -------

        """
        if self._oproject and self._odesign:
            return True
        else:
            return False

    @property
    def personallib(self):
        """Property

        :return: Full absolute path of the PersonalLib directory

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.normpath(self._desktop.GetPersonalLibDirectory())

    @property
    def userlib(self):
        """Property

        :return: Full absolute path of the UserLib directory

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.normpath(self._desktop.GetUserLibDirectory())

    @property
    def syslib(self):
        """Property

        :return: Full absolute path of the SysLib directory

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.normpath(self._desktop.GetLibraryDirectory())

    @property
    def src_dir(self):
        """Property

        :return: Full absolute path of the python directory

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.dirname(os.path.realpath(__file__))

    @property
    def pyaedt_dir(self):
        """Property

        :return: Full absolute path of the pyaedt Root Parent

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.realpath(os.path.join(self.src_dir, '..'))

    @property
    def library_list(self):
        """Property

        :return: list of [syslub, userlib, personallib]

        Parameters
        ----------

        Returns
        -------

        """
        return [self.syslib, self.userlib, self.personallib]

    @property
    def temp_directory(self):
        """Property

        :return: Full absolute path of the TEMP directory

        Parameters
        ----------

        Returns
        -------

        """
        return os.path.normpath(self._desktop.GetTempDirectory())

    @property
    def toolkit_directory(self):
        """Property

        :return: Full absolute path of the toolkit directory for this project - creates it if does not exist

        Parameters
        ----------

        Returns
        -------

        """
        toolkit_directory = os.path.join(self.project_path, self.project_name + '.toolkit')
        if not os.path.isdir(toolkit_directory):
            os.mkdir(toolkit_directory)
        return toolkit_directory

    @property
    def working_directory(self):
        """Property

        :return: Full absolute path of the working directory for this project - creates it if does not exist

        Parameters
        ----------

        Returns
        -------

        """
        working_directory = os.path.join(self.toolkit_directory, self.design_name)
        if not os.path.isdir(working_directory):
            os.mkdir(working_directory)
        return working_directory

    @property
    def default_solution_type(self):
        """Property

        :return: Default solution type for the current application running

        Parameters
        ----------

        Returns
        -------

        """
        return design_solutions[self._design_type][0]

    @property
    def odesign(self):
        """Property

        :return: oDesign object

        Parameters
        ----------

        Returns
        -------

        """
        return self._odesign

    @odesign.setter
    def odesign(self, des_name):
        """

        Parameters
        ----------
        des_name :


        Returns
        -------

        """
        warning_msg = None
        activedes = des_name
        if des_name:
            if self._assert_consistent_design_type(des_name) == des_name:
                self._insert_design(self._design_type, design_name=des_name, solution_type=self._solution_type)
        else:
            # self._odesign = self._oproject.GetActiveDesign()
            if self.design_list:
                self._odesign = self._oproject.GetDesign(self.design_list[0])
                if not self._check_design_consistency():
                    count_consistent_designs = 0
                    for des in self.design_list:
                        self._odesign = self._oproject.SetActiveDesign(des)
                        if self._check_design_consistency():
                            count_consistent_designs += 1
                            activedes = des
                    if count_consistent_designs != 1:
                        warning_msg = "No consistent unique design present - inserting a new design"
                    else:
                        self._odesign = self.oproject.SetActiveDesign(activedes)
            else:
                warning_msg = "No design present - inserting a new design"

            if warning_msg:
                self.logger.debug(warning_msg)
                self._insert_design(self._design_type, solution_type=self._solution_type)
        self.boundaries = self._get_boundaries_data()

    @property
    def oboundary(self):
        """Property

        :return: BoundarySetup Module object

        Parameters
        ----------

        Returns
        -------

        """
        return self._odesign.GetModule("BoundarySetup")

    @property
    def omodelsetup(self):
        """Property

        :return: ModelSetup Module object

        Parameters
        ----------

        Returns
        -------

        """
        return self._odesign.GetModule("ModelSetup")


    @property
    def oimportexport(self):
        """Property

        :return: ImportExport Module object

        Parameters
        ----------

        Returns
        -------

        """
        return self.odesktop.GetTool("ImportExport")

    @property
    def oproject(self):
        """Property

        :return: oProject object

        Parameters
        ----------

        Returns
        -------

        """
        return self._oproject

    @oproject.setter
    def oproject(self, proj_name=None):
        """

        Parameters
        ----------
        proj_name :
             (Default value = None)

        Returns
        -------

        """
        if not proj_name:
            self._oproject = self._desktop.GetActiveProject()
        else:
            if os.path.exists(proj_name):
                if ".aedtz" in proj_name:
                    name = self._generate_unique_project_name()

                    path = os.path.dirname(proj_name)
                    self._desktop.RestoreProjectArchive(proj_name, os.path.join(path, name), True, True)
                    time.sleep(0.5)
                    proj = self._desktop.GetActiveProject()
                elif ".def" in proj_name:
                    oTool = self._desktop.GetTool("ImportExport")
                    oTool.ImportEDB(proj_name)
                    proj = self._desktop.GetActiveProject()
                    proj.Save()
                else:
                    assert not os.path.exists(
                        proj_name + ".lock"), "Project Is Locked. Close or remove the lock before proceeding"
                    proj = self._desktop.OpenProject(proj_name)
                    time.sleep(0.5)
                self._oproject = proj
            elif proj_name in self._desktop.GetProjectList():
                self._oproject = self._desktop.SetActiveProject(proj_name)
            else:
                self._oproject = self._desktop.NewProject()
                self._oproject.Rename(os.path.join(self.project_path, proj_name+".aedt"), True)

        if not self._oproject:
            self._oproject = self._desktop.NewProject()

    @property
    def oanalysis_setup(self):
        """Property

        :return: AnalysisSetup Module object

        Parameters
        ----------

        Returns
        -------

        """
        return self.odesign.GetModule("AnalysisSetup")

    @property
    def odesktop(self):
        """Property

        :return: oDesktop Module object

        Parameters
        ----------

        Returns
        -------

        """
        return self._desktop

    @property
    def desktop_install_dir(self):
        """Property

        :return: AEDT Install Dir

        Parameters
        ----------

        Returns
        -------

        """
        return self._desktop_install_dir

    @property
    def messenger(self):
        """Property

        :return: Messenger object that can be used for logging on log file and on AEDT Message Windows

        Parameters
        ----------

        Returns
        -------

        """
        return self._messenger

    @property
    def variable_manager(self):
        """Property

        :return: Variable maanager that can be used to create and manage Project, Design and PostProcessing Variables

        Parameters
        ----------

        Returns
        -------

        """
        return self._variable_manager

    @aedt_exception_handler
    def _optimetrics_variable_args(self, arg, optimetrics_type, variable_name, min_val=None, max_val=None, tolerance=None, probability=None, mean=None, enable=True):
        """

        Parameters
        ----------
        arg :

        optimetrics_type :

        variable_name :

        min_val :
             (Default value = None)
        max_val :
             (Default value = None)
        tolerance :
             (Default value = None)
        probability :
             (Default value = None)
        mean :
             (Default value = None)
        enable :
             (Default value = True)

        Returns
        -------

        """
        if "$" in variable_name:
            tab = "NAME:ProjectVariableTab"
            propserver = "ProjectVariables"
        else:
            tab = "NAME:LocalVariableTab"
            propserver = "LocalVariables"
        arg2 = ["NAME:"+optimetrics_type, "Included:=", enable]
        if min_val:
            arg2.append("Min:=")
        arg2.append(min_val)
        if max_val:
            arg2.append("Max:=")
        arg2.append(max_val)
        if tolerance:
            arg2.append("Tol:=")
        arg2.append(tolerance)
        if probability:
            arg2.append("Prob:=")
        arg2.append(probability)
        if mean:
            arg2.append("Mean:=")
        arg2.append(mean)
        arg3 = [tab, ["NAME:PropServers", propserver], ["NAME:ChangedProps", ["NAME:" + variable_name, arg2]]]
        arg.append(arg3)

    @aedt_exception_handler
    def activate_variable_statistical(self, variable_name, min_val=None, max_val=None, tolerance=None, probability=None, mean=None):
        """Activate Statitistical Analysis for variable selected. Optionally setup ranges

        Parameters
        ----------
        variable_name :
            name of the variable
        min_val :
            Minimum value for variable. Optional (Default value = None)
        max_val :
            Maximum value for variable. Optional (Default value = None)
        tolerance :
            Tolerance value for variable. Optional (Default value = None)
        probability :
            Probability value for variable. Optional (Default value = None)
        mean :
            Mean value for variable. Optional (Default value = None)

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:AllTabs"]
        self._optimetrics_variable_args(arg, "Statistical", variable_name, min_val, max_val, tolerance, probability, mean)
        if "$" in variable_name:
            self.oproject.ChangeProperty(arg)
        else:
            self.odesign.ChangeProperty(arg)
        return True

    @aedt_exception_handler
    def activate_variable_optimization(self, variable_name, min_val=None, max_val=None):
        """Activate Optimization Analysis for variable selected. Optionally setup ranges

        Parameters
        ----------
        variable_name :
            name of the variable
        min_val :
            Minimum value for variable. Optional (Default value = None)
        max_val :
            Maximum value for variable. Optional (Default value = None)

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:AllTabs"]
        self._optimetrics_variable_args(arg, "Optimization",variable_name, min_val, max_val)
        if "$" in variable_name:
            self.oproject.ChangeProperty(arg)
        else:
            self.odesign.ChangeProperty(arg)
        return True

    @aedt_exception_handler
    def activate_variable_sensitivity(self, variable_name, min_val=None, max_val=None):
        """Activate Sensitivity Analysis for variable selected. Optionally setup ranges

        Parameters
        ----------
        variable_name :
            name of the variable
        min_val :
            Minimum value for variable. Optional (Default value = None)
        max_val :
            Maximum value for variable. Optional (Default value = None)

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:AllTabs"]
        self._optimetrics_variable_args(arg, "Sensitivity",variable_name, min_val, max_val)
        if "$" in variable_name:
            self.oproject.ChangeProperty(arg)
        else:
            self.odesign.ChangeProperty(arg)
        return True

    @aedt_exception_handler
    def activate_variable_tuning(self, variable_name, min_val=None, max_val=None):
        """Activate Tuning Analysis for variable selected. Optionally setup ranges

        Parameters
        ----------
        variable_name :
            name of the variable
        min_val :
            Minimum value for variable. Optional (Default value = None)
        max_val :
            Maximum value for variable. Optional (Default value = None)

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:AllTabs"]
        self._optimetrics_variable_args(arg, "Tuning",variable_name, min_val, max_val)
        if "$" in variable_name:
            self.oproject.ChangeProperty(arg)
        else:
            self.odesign.ChangeProperty(arg)
        return True

    @aedt_exception_handler
    def deactivate_variable_statistical(self, variable_name):
        """Deactivate Statistical Analysis for variable selected.

        Parameters
        ----------
        variable_name :
            name of the variable

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:AllTabs"]
        self._optimetrics_variable_args(arg, "Statistical",variable_name, enable=False)
        if "$" in variable_name:
            self.oproject.ChangeProperty(arg)
        else:
            self.odesign.ChangeProperty(arg)
        return True

    @aedt_exception_handler
    def deactivate_variable_optimization(self, variable_name):
        """Deactivate Optimization Analysis for variable selected.

        Parameters
        ----------
        variable_name :
            name of the variable

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:AllTabs"]
        self._optimetrics_variable_args(arg, "Optimization", variable_name, enable=False)
        if "$" in variable_name:
            self.oproject.ChangeProperty(arg)
        else:
            self.odesign.ChangeProperty(arg)
        return True

    @aedt_exception_handler
    def deactivate_variable_sensitivity(self, variable_name):
        """Deactivate Sensitivity Analysis for variable selected.

        Parameters
        ----------
        variable_name :
            name of the variable

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:AllTabs"]
        self._optimetrics_variable_args(arg, "Sensitivity",variable_name, enable=False)
        if "$" in variable_name:
            self.oproject.ChangeProperty(arg)
        else:
            self.odesign.ChangeProperty(arg)
        return True

    @aedt_exception_handler
    def deactivate_variable_tuning(self, variable_name):
        """Deactivate Tuning Analysis for variable selected.

        Parameters
        ----------
        variable_name :
            name of the variable

        Returns
        -------
        type
            Bool

        """
        arg = ["NAME:AllTabs"]
        self._optimetrics_variable_args(arg, "Tuning",variable_name, enable=False)
        if "$" in variable_name:
            self.oproject.ChangeProperty(arg)
        else:
            self.odesign.ChangeProperty(arg)
        return True

    @aedt_exception_handler
    def _get_boundaries_data(self):
        """ """
        boundaries = []
        if self.design_properties and 'BoundarySetup' in self.design_properties:
            for ds in self.design_properties['BoundarySetup']["Boundaries"]:
                try:
                    if type(self.design_properties['BoundarySetup']["Boundaries"][ds]) is OrderedDict:
                        boundaries.append(BoundaryObject(self, ds,
                                                        self.design_properties['BoundarySetup']["Boundaries"][ds],
                                                        self.design_properties['BoundarySetup']["Boundaries"][ds][
                                                            'BoundType']))
                except:
                    pass
        return boundaries


    @aedt_exception_handler
    def _get_ds_data(self, name, datas):
        """

        Parameters
        ----------
        name :

        datas :


        Returns
        -------

        """
        units = datas["DimUnits"]
        numcol = len(units)
        x = []
        y = []
        z = None
        v = None
        if numcol > 2:
            z = []
            v = []
        if 'Coordinate' in datas:
            for el in datas['Coordinate']:
                x.append(el['CoordPoint'][0])
                y.append(el['CoordPoint'][1])
                if numcol > 2:
                    z.append(el['CoordPoint'][2])
                    v.append(el['CoordPoint'][3])
        else:
            new_list = [datas['Points'][i:i + numcol] for i in range(0, len(datas['Points']), numcol)]
            for el in new_list:
                x.append(el[0])
                y.append(el[1])
                if numcol > 2:
                    z.append(el[2])
                    v.append(el[3])
        if numcol == 2:
            return DataSet(self, name, x, y, z, v, units[0], units[1])
        else:
            return DataSet(self, name, x, y, z, v, units[0], units[1], units[2], units[3])

    @aedt_exception_handler
    def _get_project_datasets(self):
        """ """
        datasets = {}
        try:
            for ds in self.project_properies['AnsoftProject']['ProjectDatasets']['DatasetDefinitions']:
                datas = self.project_properies['AnsoftProject']['ProjectDatasets']['DatasetDefinitions'][ds]['Coordinates']
                datasets[ds] = self._get_ds_data(ds, datas)
        except:
            pass
        return datasets


    @aedt_exception_handler
    def _get_design_datasets(self):
        """ """
        datasets = {}
        try:
            for ds in self.design_properties['ModelSetup']['DesignDatasets']['DatasetDefinitions']:
                datas = self.project_properies['ModelSetup']['DesignDatasets']['DatasetDefinitions'][ds]['Coordinates']
                datasets[ds] = self._get_ds_data(ds, datas)
        except:
            pass
        return datasets

    @aedt_exception_handler
    def close_desktop(self):
        """:return: Close the Desktop and release the tool"""
        force_close_desktop()
        return True

    @aedt_exception_handler
    def release_desktop(self, close_projects=True, close_desktop=True):
        """

        Parameters
        ----------
        close_projects: bool
            close all projects
        close_desktop: bool
            close desktop after release it

        Returns
        -------
        bool
            True if desktop released
        """
        release_desktop(close_projects, close_desktop)
        return True

    @aedt_exception_handler
    def generate_temp_project_directory(self, subdir_name):
        """Generate a unique directory string to store a project

        Creates a directory name for storage of a project in a location which is guaranteed to exist (the temp
        directory of the AEDT installation). If the 'name' parameter is defined, then add a sub-directory within
        the temp directory and add a hash suffix to ensure that this directory is empty (unique name)

        Parameters
        ----------
        subdir_name : str
            Base name of the sub-directory to be created im self.tempdirectory

        Returns
        -------
        str

        Examples
        --------
        >>> m3d = Maxwell3d()
        >>> proj_directory = m3d.generate_temp_project_directory("Example")

        """
        base_path = self.temp_directory

        if not isinstance(subdir_name, str):
            self.messenger.add_error_message("Input argument 'subdir' must be a string")
            return False
        dir_name = generate_unique_name(subdir_name)
        project_dir = os.path.join(base_path, dir_name)
        try:
            if not os.path.exists(project_dir): os.makedirs(project_dir)
            return project_dir
        except OSError:
            return False


    @aedt_exception_handler
    def load_project(self, project_file, design_name=None, close_active_proj=False):
        """Open aedt project based on project file. Design is optional

        Parameters
        ----------
        project_file :
            full path project file name
        design_name :
            optional design name (Default value = None)
        close_active_proj :
            bool (Default value = False)

        Returns
        -------
        type
            bool

        """
        if close_active_proj:
            self._desktop.CloseProject(self.project_name)
        proj = self._desktop.OpenProject(project_file)
        time.sleep(0.5)
        self._odesign = None
        self._oproject = None
        self.oproject = proj.GetName()
        time.sleep(0.5)
        self.odesign = design_name
        time.sleep(0.5)
        if proj:
            return True
        else:
            return False

    @aedt_exception_handler
    def create_dataset1d_design(self, dsname, xlist, ylist, xunit="", yunit=""):
        """Create a Design dataset

        Parameters
        ----------
        dsname :
            name of dataset (without prefix for project datasets)
        xlist :
            list of x values of dataset
        ylist :
            list of y values of dataset
        xunit :
            optional x unit (Default value = "")
        yunit :
            optional y unit (Default value = "")

        Returns
        -------
        type
            Dataset object if dataset is created. False if not

        """
        return self.create_dataset(dsname, xlist,ylist, is_project_dataset=False, xunit=xunit, yunit=yunit)

    @aedt_exception_handler
    def create_dataset1d_project(self, dsname, xlist, ylist, xunit="", yunit=""):
        """Create a Project dataset

        Parameters
        ----------
        dsname :
            name of dataset (without prefix for project datasets)
        xlist :
            list of x values of dataset
        ylist :
            list of y values of dataset
        xunit :
            optional x unit (Default value = "")
        yunit :
            optional y unit (Default value = "")

        Returns
        -------
        type
            Dataset object if dataset is created. False if not

        """
        return self.create_dataset(dsname, xlist,ylist, is_project_dataset=True, xunit=xunit, yunit=yunit)

    @aedt_exception_handler
    def create_dataset3d(self, dsname, xlist, ylist, zlist=None, vlist=None, xunit="", yunit="",
                       zunit="", vunit=""):
        """Create a Dataset

        Parameters
        ----------
        dsname :
            name of dataset (without prefix for project datasets)
        xlist :
            list of x values of dataset
        ylist :
            list of y values of dataset
        zlist :
            list of z values of dataset (for 3D Datasets only) (Default value = None)
        vlist :
            list of v values of dataset (for 3D Datasets only) (Default value = None)
        xunit :
            optional x unit (Default value = "")
        yunit :
            optional y unit (Default value = "")
        zunit :
            optional z unit (for 3D Datasets only) (Default value = "")
        vunit :
            optional v unit (for 3D Datasets only) (Default value = "")

        Returns
        -------
        type
            Dataset object if dataset is created. False if not

        """
        return self.create_dataset(dsname=dsname, xlist=xlist, ylist=ylist, zlist=zlist, vlist=vlist, xunit=xunit,
                                   yunit=yunit, zunit=zunit, vunit=vunit)

    @aedt_exception_handler
    def create_dataset(self, dsname, xlist, ylist, zlist=None, vlist=None, is_project_dataset=True, xunit="", yunit="",
                       zunit="", vunit=""):
        """Create a Dataset

        Parameters
        ----------
        dsname :
            name of dataset (without prefix for project datasets)
        xlist :
            list of x values of dataset
        ylist :
            list of y values of dataset
        zlist :
            list of z values of dataset (for 3D Datasets only) (Default value = None)
        vlist :
            list of v values of dataset (for 3D Datasets only) (Default value = None)
        is_project_dataset :
            bool True if is a project dataset, False if it is a design dataset (Default value = True)
        xunit :
            optional x unit (Default value = "")
        yunit :
            optional y unit (Default value = "")
        zunit :
            optional z unit (for 3D Datasets only) (Default value = "")
        vunit :
            optional v unit (for 3D Datasets only) (Default value = "")

        Returns
        -------
        type
            Dataset object if dataset is created. False if not

        """
        if not self.dataset_exists(dsname, is_project_dataset):
            if is_project_dataset:
                dsname = "$" + dsname
            ds = DataSet(self, dsname, xlist, ylist, zlist, vlist, xunit, yunit, zunit, vunit)
        else:
            self.messenger.add_warning_message("Dataset {} already exists".format(dsname))
            return False
        ds.create()
        if is_project_dataset:
            self.project_datasets[dsname] = ds
        else:
            self.design_datasets[dsname] = ds
        return ds

    @aedt_exception_handler
    def dataset_exists(self, name, is_project_dataset=True):
        """Check if a dataset exists

        Parameters
        ----------
        name :
            Name of dataset to check  (without prefix for project datasets)
        is_project_dataset :
            bool (Default value = True)

        Returns
        -------
        type
            True if dataset exists

        """
        if is_project_dataset and "$"+name in self.project_datasets:
            self.messenger.add_info_message("DATASET {} exists.".format("$"+name))
            return True
            #self.oproject.ExportDataSet("$"+name, os.path.join(self.temp_directory, "ds.tab"))
        elif not is_project_dataset and name in self.design_datasets:
            self.messenger.add_info_message("DATASET {} exists.".format(name))
            return True
            #self.odesign.ExportDataSet(name, os.path.join(self.temp_directory, "ds.tab"))
        self.messenger.add_info_message("DATASET {} doesn't exists.".format(name))
        return False


    @aedt_exception_handler
    def change_automatically_use_causal_materials(self, lossy_dielectric=True):
        """Enable or disable the automatic use of causal material for lossy dielectrics

        Parameters
        ----------
        lossy_dielectric : bool
            enable or disable causal materials (Default value = True)

        Returns
        -------

        """
        if lossy_dielectric:
            self._messenger.add_info_message("Enabling Automatic use of causal materials")
        else:
            self._messenger.add_info_message("Disabling Automatic use of causal materials")
        self.odesign.SetDesignSettings(
            ["NAME:Design Settings Data", "Calculate Lossy Dielectrics:=", lossy_dielectric])
        return True

    @aedt_exception_handler
    def change_material_override(self, material_override=True):
        """Enable the material override in the given Project

        Parameters
        ----------
        ___________ :

        material_override :
            bool (Default value = True)
        enable :
            or disable material override

        Returns
        -------
        type
            _______
            bool

        """
        if material_override:
            self._messenger.add_info_message("Enabling Material Override")
        else:
            self._messenger.add_info_message("Disabling Material Override")
        self.odesign.SetDesignSettings(
            ["NAME:Design Settings Data", "Allow Material Override:=", material_override])
        return True

    @aedt_exception_handler
    def change_validation_settings(self, entity_check_level="Strict", ignore_unclassified=False,
                                   skip_intersections=False):
        """change the validation design settings

        Parameters
        ----------
        entity_check_level : str
            Entity Level, default "Strict"
        ignore_unclassified : bool
            Ignore unclassified elements (Default value = False)
        skip_intersections : bool
            Skip Intersections (Default value = False)

        Returns
        -------

            bool

        """
        self._messenger.add_info_message("Changing the validation design settings")
        self.odesign.SetDesignSettings(["NAME:Design Settings Data"],
                                       ["NAME:Model Validation Settings",
                                        "EntityCheckLevel:=", entity_check_level,
                                        "IgnoreUnclassifiedObjects:=", ignore_unclassified,
                                        "SkipIntersectionChecks:=", skip_intersections])
        return True

    @aedt_exception_handler
    def clean_proj_folder(self, directory=None, name=None):
        """Delete all project name related folder

        Parameters
        ----------
        directory : str, default=None
            project directory, of not specified clear the .aedtresults directory of the active project
        name : str, default=None
            project name, if not specified clear the data for the active project

        Returns
        -------
        bool
            True if the clear operation is successful
        """
        if not name:
            name=self.project_name
        if not directory:
            directory = self.results_directory
        self._messenger.add_info_message("Cleanup folder {} from project {}".format(directory, name))
        if os.path.exists(directory):
            shutil.rmtree(directory, True)
            if not os.path.exists(directory):
                os.mkdir(directory)
        self._messenger.add_info_message("Project Directory cleaned")
        return True

    @aedt_exception_handler
    def copy_project(self, path, dest):
        """Copy the given project to another destination
        Save the project before copying it

        Parameters
        ----------
        path :
            destination project path
        dest :
            name of the project

        Returns
        -------

        """
        self._messenger.add_info_message("Copy AEDT Project ")
        self.oproject.Save()
        self.oproject.SaveAs(os.path.join(path, dest + ".aedt"), True)
        return True

    @aedt_exception_handler
    def create_new_project(self, proj_name):
        """Create a new project within the desktop

        Parameters
        ----------
        proj_name :
            name of the new project

        Returns
        -------

        """
        self._messenger.add_info_message("Creating new Project ")
        prj = self._desktop.NewProject(proj_name)
        prj_name = prj.GetName()
        self.oproject = prj_name
        self.odesign = None
        return True

    @aedt_exception_handler
    def close_project(self, name=None, saveproject=True):
        """Close the specified project

        Parameters
        ----------
        name :
            name of the project (if none take active project) (Default value = None)
        saveproject : bool
            Save project before close (Default value = True)

        Returns
        -------

        """
        msg_txt = ""
        if name:
            assert name in self.project_list, "Invalid project name {}".format(name)
            msg_txt = "specified "+ name
        else:
            name = self.project_name
            msg_txt = "active "+ self.project_name
        self._messenger.add_info_message("Closing the {} AEDT Project".format(msg_txt), level="Global")
        if name != self.project_name:
            oproj = self.odesktop.SetActiveProject(name)
        else:
            oproj = self.oproject
        if saveproject:
            oproj.Save()

        lock_file = str(self.lock_file)
        self.odesktop.CloseProject(name)
        if name == self.project_name:
            assert not os.path.exists(lock_file), 'AEDT project did not close properly'
        return True

    @aedt_exception_handler
    def delete_design(self, name=None, fallback_design=None):
        """Delete name Design from the current project (will not work from toolkit)

        Parameters
        ----------
        name : str, default=None
            Design name to be deleted. If not specified then delete the active design

        Returns
        -------
        bool
            succeed status
        """
        if not name:
            name = self.design_name
        self.oproject.DeleteDesign(name)
        try:
            if fallback_design:
                new_designname = fallback_design
            else:
                if not self._oproject.GetActiveDesign():
                    new_designname = None
                    self.odesign = None
                else:
                    new_designname = self._oproject.GetActiveDesign().GetName()
                    self.set_active_design(new_designname)
        except:
            pass
        self.__init__(self.project_name, self.design_name)

        return True


    @aedt_exception_handler
    def delete_separator(self, separator_name):
        """Deletes a separator from either the active project or design

        Parameters
        ----------
        separator_name :


        Returns
        -------
        bool
            True if the separator exists and can be deleted, False otherwise

        """
        return self._variable_manager.delete_separator(separator_name)

    @aedt_exception_handler
    def delete_variable(self, sVarName):
        """Delete a Variable

        Parameters
        ----------
        sVarName :
            name of variable to delete

        Returns
        -------
        type
            none

        """
        return  self.variable_manager.delete_variable(sVarName)

    @aedt_exception_handler
    def insert_design(self, design_name=None):
        """Inserts a design of the specified design type. Default design type is taked from the derived application \
        class. If no design-name is given, the default design name is <Design-Type>Design<_index>. If the given or \
        default design name is in use, then an underscore + index is added            to ensure that the design name\
        is unique. The inserted object is assigned to self._odesign

        Parameters
        ----------
        design_name :
            optional design name (Default value = None)

        Returns
        -------

        """"Circuit Design"
        if self.design_type == "Circuit Design" or self.design_type == "HFSS 3D Layout Design":
            self.modeler.edb.close_edb()
        self.__init__(projectname=self.project_name, designname=design_name)

    def _insert_design(self, design_type, design_name=None, solution_type=None):
        assert design_type in design_solutions, "Invalid design type for insert: {}".format(design_type)
        # self.save_project() ## Commented because it saves a Projectxxx.aedt when launched on an empty Desktop
        unique_design_name = self._generate_unique_design_name(design_name)
        if solution_type:
            assert solution_type in design_solutions[self._design_type], \
                "Invalid solution type {0} for design type {1}".format(solution_type, self._design_type)
        else:
            solution_type = self.default_solution_type
        if design_type == "RMxprtSolution":
            new_design = self._oproject.InsertDesign("RMxprt", unique_design_name, "Inner-Rotor Induction Machine", "")
        elif design_type == "ModelCreation":
            new_design = self._oproject.InsertDesign("RMxprt", unique_design_name, "Model Creation Inner-Rotor Induction Machine", "")
        else:
            new_design = self._oproject.InsertDesign(design_type, unique_design_name, solution_type, "")
        self._messenger.add_info_message("Added design '{0}' of type {1}".format(unique_design_name, design_type),
                                         level='Project')
        name = new_design.GetName()
        if ";" in name:
            self.odesign = name.split(";")[1]
        else:
            self.odesign = name
        return name

    @aedt_exception_handler
    def _generate_unique_design_name(self, design_name):
        """

        Parameters
        ----------
        design_name :


        Returns
        -------

        """
        design_index = 0
        suffix = ""
        if not design_name:
            char_set = string.ascii_uppercase + string.digits
            uName = ''.join(random.sample(char_set, 3))
            design_name = self._design_type + "_" + uName
        while design_name in self.design_list:
            if design_index:
                design_name = design_name[0:-len(suffix)]
            design_index += 1
            suffix = "_" + str(design_index)
            design_name += suffix
        return design_name

    @aedt_exception_handler
    def _generate_unique_project_name(self):
        """ """
        char_set = string.ascii_uppercase + string.digits
        uName = ''.join(random.sample(char_set, 3))
        proj_name = "Project_" + uName+ ".aedt"
        return proj_name

    @aedt_exception_handler
    def rename_design(self, new_name):
        """Rename the active Design

        Parameters
        ----------
        new_name :
            new name of the design

        Returns
        -------

        """
        self._odesign.RenameDesignInstance(self.design_name, new_name)
        self.odesign = new_name
        return True

    @aedt_exception_handler
    def copy_design_from(self, project_fullname, design_name):
        """Copy the design with name design_name from the project project_fullname  into the active design\
        If a design with the same name is already present in the destination project, the name is\
         automatically changed (this is done by AEDT). The name is returned by the function.\
        The active design is maintained.

        Parameters
        ----------
        project_fullname :
            copmlete full name (with path) of the project containing the design to be copied
        design_name :
            design name to be copied into the active design

        Returns
        -------
        type
            the copied design name or None if not copied (e.g. design_name does not exists in project_fullname)

        """
        self.save_project()
        active_design = self.design_name
        # open the origin project
        if os.path.exists(project_fullname):
            proj_from = self._desktop.OpenProject(project_fullname)
            proj_from_name = proj_from.GetName()
        else:
            return None
        # check if the requested design exists in the origin project
        if design_name not in [i.GetName() for i in list(proj_from.GetDesigns())]:
            return None
        # copy the source design
        proj_from.CopyDesign(design_name)
        # paste in the destination project and get the name
        self._oproject.Paste()
        new_designname = self._oproject.GetActiveDesign().GetName()
        if self._oproject.GetActiveDesign().GetDesignType() == 'HFSS 3D Layout Design':
            new_designname = new_designname[2:]  # name is returned as '2;EMDesign3'
        # close the source project
        self._desktop.CloseProject(proj_from_name)
        # reset the active design (very important)
        self._oproject.SetActiveDesign(active_design)
        self.save_project()
        self.__init__(self.project_name, new_designname)

        # return the pasted design name
        return new_designname

    @aedt_exception_handler
    def duplicate_design(self, label):
        """Copy the selected design with a new name consisting of the original design name, plus
            a defined extension MMode, plus a running index as necessary to allow for multiple calls

        Parameters
        ----------
        label :


        Returns
        -------

        """

        active_design = self.design_name
        design_list = self.design_list
        self._oproject.CopyDesign(active_design)
        self._oproject.Paste()
        newname = label
        ind = 1
        while newname in self.design_list:
            newname = label + '_' + str(ind)
            ind += 1
        actual_name = [i for i in self.design_list if i not in design_list]
        self.odesign = actual_name
        self.design_name = newname
        self.__init__(self.project_name, self.design_name)

        return True

    @aedt_exception_handler
    def export_variables_to_csv(self, filename, export_project=True, export_design=True):
        """Export Project and design variables variables to filename

        Parameters
        ----------
        filename :
            full path to csv file to export data
        export_design :
            True (Export Design Variables) | False (Default value = True)
        export_project :
            True (Export Project Variables) | False (Default value = True)

        Returns
        -------
        type
            True

        """
        """
            
        """
        varnames = []
        desnames = []
        if export_project:
            varnames = self.oproject.GetProperties("ProjectVariableTab", "ProjectVariables")
        if export_design:
            desnames = self.odesign.GetProperties("LocalVariableTab", "LocalVariables")
        with open(filename, 'w') as csvfile:
            filewriter = csv.writer(csvfile, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
            filewriter.writerow(['Name', 'Value'])
            for el in varnames:
                value = self.oproject.GetVariableValue(el)
                filewriter.writerow([el, value])
            for el in desnames:
                value = self.odesign.GetVariableValue(el)
                filewriter.writerow([el, value])
        return True

    @aedt_exception_handler
    def read_design_data(self):
        """Reads back the design data as a dictionary"""
        design_file = os.path.join(self.working_directory, "design_data.json")
        with open(design_file, 'r') as fps:
            design_data = json.load(fps)
        return design_data

    @aedt_exception_handler
    def save_project(self, project_file=None, overwrite=True, refresh_obj_ids_after_save=False):
        """Save the AEDT Project and add a message.

        Parameters
        ----------
        project_file : str, optional
            Full Project path. Default ``None``.
        overwrite : bool, optional
            Overwrite existing project.  Default ``True``.
        refresh_obj_ids_after_save : bool
            Refresh object IDs after save.  Default ``False``.

        Returns
        -------
        bool
            Always returns ``True``

        """
        msg_text = "Saving {0} Project".format(self.project_name)
        self._messenger.add_info_message(msg_text, level='Global')
        if project_file:
            self.oproject.SaveAs(project_file, overwrite)
        else:
            self.oproject.Save()
        if refresh_obj_ids_after_save:
            self.modeler.primitives._refresh_all_ids_from_aedt_file()
        return True

    @aedt_exception_handler
    def delete_project(self, project_name):
        """Delete the named project.

        Parameters
        ----------
        project_name : str
            Name of the project to delete

        Returns
        -------
        bool
            Always returns ``True``.

        """
        assert self.project_name != project_name, "Cannot delete active design"
        self._desktop.DeleteProject(project_name)
        return True

    @aedt_exception_handler
    def set_active_design(self, name):
        """Change active Design to another design

        Parameters
        ----------
        name :
            design name to make active

        Returns
        -------
        type
            nothing

        """
        self.oproject.SetActiveDesign(name)
        self.odesign = name
        self.__init__(self.project_name, self.design_name)
        return True

    @aedt_exception_handler
    def update_registry_from_file(self, filename):
        """Update HPC options from an af from an acf configuration file.

        Parameters
        ----------
        filename :
            full path to the filename (can be an acf or a txt)

        Returns
        -------
        type
            nothing

        """
        self._desktop.SetRegistryFromFile(r'%s' % os.path.abspath(filename))
        return True

    @aedt_exception_handler
    def validate_simple(self, logfile=None):
        """Validate Design. Return 0 if validation failed or 1 if passed

        Parameters
        ----------
        logfile :
            Optional, save validation to logfile (Default value = None)

        Returns
        -------

        """
        if logfile:
            return self._odesign.ValidateDesign(logfile)
        else:
            return self._odesign.ValidateDesign()


    @aedt_exception_handler
    def get_evaluated_value(self, variable_name, variation=None):
        """Returns the floating-point evaluated value of a given variable in SI-units
        (design property or project variable)

        Optionally a variation string of the form

        Parameters
        ----------
        variable_name : str, default=None
        Name of the project- or design variable to be evaluated
        variation : str, default=None
        Variation string for the evaluation. If not specified then use the nominal variation

        >>> M3D = Maxwell3D()
        >>> M3D["p1"] = "10mm"
        >>> M3D["p2"] = "20mm"
        >>> M3D["p3"] = "P1 * p2"
        >>> eval_p3 = M3D.get_evaluated_value("p3")

        Returns
        -------
        float

        """
        if not variation:
            variation_string = self._odesign.GetNominalVariation()
        else:
            variation_string = self.design_variation(variation_string=variation)

        si_value = self._odesign.GetVariationVariableValue(variation_string, variable_name)

        return si_value

    @aedt_exception_handler
    def evaluate_expression(self, expression_string):
        """ Evaluate an arbitrary valid string expression and return the numerical value in SI units

        Parameters
        ----------
        expression_string : str
        A valid design property/project variable string, i.e "34mm*sqrt(2)", "$G1*p2/34"

        Returns
        -------
        float

        """
        # Set the value of an internal reserved design variable to the specified string
        if expression_string in self._variable_manager.variables:
            return self._variable_manager.variables[expression_string]
        else:
            try:
                self._variable_manager.set_variable("pyaedt_evaluator", expression=expression_string,
                                                    readonly=True, hidden=True, description="Internal_Evaluator")
            except:
                raise("Invalid string expression {}".expression_string)

            # Extract the numeric value of the expression (in SI units!)
            return self._variable_manager.variables["pyaedt_evaluator"].value

    @aedt_exception_handler
    def design_variation(self, variation_string=None):
        """Generate a string to specify a desired variation

        Converts a user input string defining a desired solution variation, e.g. "p1=1mm p2=3mm" into a valid
        string taking into account all existing design properties and project variables including non-sweep
        dependent properties.

        This is actually needed because the standard GetVariationVariableValue does not work for obtaining
        values of dependent (non-sweep variables). Presumably using the new beta feature object-oriented
        scripting model this could be made redundant in future releases.

        Parameters
        ----------
        variation_string : str
        Variation string of the form "p1=1mm p2=3mm"

        Returns
        -------
        str

        """
        nominal = self._odesign.GetNominalVariation()
        if variation_string:
            # decompose the nominal variation into a dictionary of name[value]
            nominal_dict = variation_string_to_dict(nominal)

            # decompose the desired variation into a dictionary of name[value]
            var_dict = variation_string_to_dict(variation_string)

            # set the values of the desired variation in the active design
            for key, value in var_dict.items():
                self[key] = value

            # get the desired variation values
            nominal = self._odesign.GetNominalVariation()

            # Reset the nominal values in the active design
            for key in var_dict:
                self[key] = nominal_dict[key]

        return nominal

    @aedt_exception_handler
    def _assert_consistent_design_type(self, des_name):
        """

        Parameters
        ----------
        des_name :


        Returns
        -------

        """
        if des_name in self.design_list:
            self._odesign = self._oproject.SetActiveDesign(des_name)
            dtype = self._odesign.GetDesignType()
            if dtype != "RMxprt":
                assert dtype == self._design_type, \
                    "Error: Specified design is not of type {}".format(self._design_type)
            else:
                assert ("RMxprtSolution" == self._design_type) or ("ModelCreation" == self._design_type), \
                    "Error: Specified design is not of type {}".format(self._design_type)
            return True
        else:
            return des_name

    @aedt_exception_handler
    def _check_solution_consistency(self):
        """ """
        if self._solution_type:
            return self._odesign.GetSolutionType() in self._solution_type
        else:
            return True

    @aedt_exception_handler
    def _check_design_consistency(self):
        """ """
        consistent = False
        destype = self._odesign.GetDesignType()
        if destype == self._design_type:
            consistent = self._check_solution_consistency()
        return consistent



import os
import string
import random
import time
import sys
import traceback
import logging
from functools import wraps
from collections import OrderedDict
import inspect
import itertools
logger = logging.getLogger(__name__)
import pkgutil
modules = [tup[1] for tup in pkgutil.iter_modules()]
if 'clr' in modules:
    import clr

class MethodNotSupportedError(Exception):
    """ """
    pass

def _exception(ex_info, func, args, kwargs, message="Type Error"):
    """Writes the trace stack to the desktop when a python error occurs

    Parameters
    ----------
    ex_info :
        
    func :
        
    args :
        
    kwargs :
        
    message :
         (Default value = "Type Error")

    Returns
    -------

    """
    print("**************************************************************")
    print("pyaedt Error on Method {}:  {}. Please Check again".format(func.__name__, message))
    print("Arguments Provided: ")
    try:

        if int(sys.version[0]) > 2:
            args_name = list(OrderedDict.fromkeys(inspect.getfullargspec(func)[0] + list(kwargs.keys())))
            args_dict = OrderedDict(list(itertools.zip_longest(args_name, args)) + list(kwargs.items()))
        else:
            args_name = list(OrderedDict.fromkeys(inspect.getargspec(func)[0] + list(kwargs.keys())))
            args_dict = OrderedDict(list(itertools.izip(args_name, args)) + list(kwargs.iteritems()))

        for el in args_dict:
            if el != "self":
                print("    {} = {} ".format(el, args_dict[el]))
    except:
        if len(args) > 1:
            print(args[1:], kwargs)
        else:
            print(kwargs)
    ex_value = ex_info[1]
    tb_data = ex_info[2]
    tb_trace = traceback.format_tb(tb_data)
    if len(tb_trace) > 1:
        tblist = tb_trace[1].split('\n')
    else:
        tblist = tb_trace[0].split('\n')
    print("")
    print(str(ex_value))
    if logger:
        logger.error(str(ex_value))
    # self._main.oDesktop.AddMessage(proj_name, des_name, 2, str(ex_value))
    for el in tblist:
        # self._main.oDesktop.AddMessage(proj_name, des_name, 2, el)
        if "inner_function" not in el and "**kwargs" not in el:
            print(el)
            if logger:
                logger.error(el)
    print("")
    print("")
    print("Method Docstring: ")
    print("")
    print(func.__doc__)
    print("************************************************************")


def aedt_exception_handler(func):
    """Decorator for pyaedt Exception Management

    Parameters
    ----------
    func :
        method to be decorated

    Returns
    -------
    type
        function return if correctly executed otherwise it will return False and errors will be plotted

    """
    @wraps(func)
    def inner_function(*args, **kwargs):
        """

        Parameters
        ----------
        *args :
            
        **kwargs :
            

        Returns
        -------

        """
        try:
            return func(*args, **kwargs)
        except TypeError:
            _exception(sys.exc_info(), func, args, kwargs, "Type Error")
            return False
        except ValueError:
            _exception(sys.exc_info(), func, args, kwargs, "Value Error")
            return False
        except AttributeError:
            _exception(sys.exc_info(), func, args, kwargs, "Attribute Error")
            return False
        except KeyError:
            _exception(sys.exc_info(), func, args, kwargs, "Key Error")
            return False
        except IndexError:
            _exception(sys.exc_info(), func, args, kwargs, "Index Error")
            return False
        except AssertionError:
            _exception(sys.exc_info(), func, args, kwargs, "Assertion Error")
            return False
        except NameError:
            _exception(sys.exc_info(), func, args, kwargs, "Name Error")
            return False
        except IOError:
            _exception(sys.exc_info(), func, args, kwargs, "IO Error")
            return False
        except MethodNotSupportedError:
            message = "This Method is not supported in current AEDT Design Type."
            print("**************************************************************")
            print("pyaedt Error on Method {}:  {}. Please Check again".format(func.__name__, message))
            print("**************************************************************")
            print("")
            logger.error(message)
            return False
        except BaseException:
            _exception(sys.exc_info(), func, args, kwargs, "General or AEDT Error")
            return False
    return inner_function


@aedt_exception_handler
def env_path(input_version):
    """

    Parameters
    ----------
    input_version :
        

    Returns
    -------

    """
    version = int(input_version[2:4])
    release = int(input_version[5])
    if version < 20:
        if release < 3:
            version -= 1
        else:
            release -= 2
    v_key = "ANSYSEM_ROOT{0}{1}".format(version, release)
    return os.getenv(v_key)


@aedt_exception_handler
def env_value(input_version):
    """

    Parameters
    ----------
    input_version :


    Returns
    -------

    """
    version = int(input_version[2:4])
    release = int(input_version[5])
    if version < 20:
        if release < 3:
            version -= 1
        else:
            release -= 2
    v_key = "ANSYSEM_ROOT{0}{1}".format(version, release)
    return v_key

@aedt_exception_handler
def get_filename_without_extension(path):
    """

    Parameters
    ----------
    path :
        

    Returns
    -------

    """
    return os.path.splitext(os.path.split(path)[1])[0]


@aedt_exception_handler
def generate_unique_name(rootname, suffix='', n=6):
    """Generate a new Random name given a rootname and, optionally a suffix

    Parameters
    ----------
    rootname :
        Root name to which add 6 Random chars
    suffix :
        Suffix to be added (Default value = '')
    n :
        Number of random characters in the name, defaults to 6

    Returns
    -------

    """
    char_set = string.ascii_uppercase + string.digits
    uName = ''.join(random.choice(char_set) for _ in range(n))
    unique_name = rootname + "_" + uName
    if suffix:
        unique_name += "_" + suffix
    return unique_name


def retry_ntimes(n, function, *args, **kwargs):
    """

    Parameters
    ----------
    n :
        
    function :
        
    *args :
        
    **kwargs :
        

    Returns
    -------

    """
    retry = 0
    ret_val = None
    while retry < n:
        try:
            ret_val = function(*args, **kwargs)
            if not ret_val and type(ret_val) is not float and type(ret_val) is not int:
                ret_val = True
        except:
            retry += 1
            time.sleep(0.05)
        else:
            break
    return ret_val




from ..generic.general_methods import aedt_exception_handler
from .Primitives import Primitives, Polyline
from .GeometryOperators import GeometryOperators
from .Object3d import Object3d
import os

class Primitives3D(Primitives, object):
    """Class for management of all Primitives of 3D Tools"""

    def __init__(self, parent, modeler):
        Primitives.__init__(self, parent, modeler)

    @aedt_exception_handler
    def is3d(self):
        """Returns True always to indicate a 3D analysis type"""
        return True

    @aedt_exception_handler
    def create_box(self, position, dimensions_list, name=None, matname=None):
        """Create a Box


        Parameters
        ----------
        position :
            ApplicationName.modeler.Position(x,y,z) object
        dimensions_list :
            list of dimensions of X, Y, Z
        name :
            box name. Optional, if nothing default name will be assigned
        matname :
            material name. Optional, if nothing or if invalid, default material will be assigned

        Returns
        -------
        int
            Box ID

        Examples
        _________
        >>> from pyaedt import hfss
        >>> hfss = HFSS()
        >>> origin = [0,0,0]
        >>> dimensions = [10,5,20]
        >>> #Material and name are not mandatory fields
        >>> object_id = hfss.modeler.primivites.create_box(origin, dimensions, name="mybox", matname="copper")
        """
        o = self._new_object(matname=matname)

        XPosition, YPosition, ZPosition = self.pos_with_arg(position)
        if XPosition is None or YPosition is None or ZPosition is None:
            raise AttributeError("Position Argument must be a valid 3 Element List")
        XSize, YSize, ZSize = self.pos_with_arg(dimensions_list)
        if XSize is None or YSize is None or YSize is None:
            raise AttributeError("Dimension Argument must be a valid 3 Element List")
        vArg1 = ["NAME:BoxParameters"]
        vArg1.append("XPosition:="), vArg1.append(XPosition)
        vArg1.append("YPosition:="), vArg1.append(YPosition)
        vArg1.append("ZPosition:="), vArg1.append(ZPosition)
        vArg1.append("XSize:="), vArg1.append(XSize)
        vArg1.append("YSize:="), vArg1.append(YSize)
        vArg1.append("ZSize:="), vArg1.append(ZSize)

        if self.version >= "2019.3":
            vArg2 = o.export_attributes(name)
        else:
            vArg2 = o.export_attributes_legacy(name)
        o._m_name = self.oeditor.CreateBox(vArg1, vArg2)

        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_bondwire(self, start_position, end_position, h1=0.2, h2=0, alpha=80, beta=5, bond_type=0,
                        diameter=0.025, facets=6, name=None, matname=None):
        """Create a Bondwire.

        Parameters
        ----------
        start_position : list
            Starting Position
        end_position :list
            Ending Position
        h1: float
            h1 value
        h2: float
            h2 value
        alpha: float
            alpha angle
        beta: float
            beta angle
        bond_type: int
            0- JEDEC5, 1- Jedec4, 2- Low. Default JEDEC_5
        diameter: float
            wire diameter
        facets: int
            wire facets
        name :
            box name. Optional, if nothing default name will be assigned
        matname :
            material name. Optional, if nothing default material will be assigned

        Returns
        -------
        int
            Box ID

        Examples
        _________
        >>> from pyaedt import Hfss
        >>> hfss = Hfss()
        >>> origin = [0,0,0]
        >>> endpos = [10,5,20]
        >>> #Material and name are not mandatory fields
        >>> object_id = hfss.modeler.primivites.create_bondwire(origin, endpos,h1=0.5, h2=0.1, alpha=75, beta=4,bond_type=0, name="mybox", matname="copper")
        """
        o = self._new_object(matname=matname)

        XPosition, YPosition, ZPosition = self.pos_with_arg(start_position)
        if XPosition is None or YPosition is None or ZPosition is None:
            raise AttributeError("Position Argument must be a valid 3 Element List")
        XSize, YSize, ZSize = self.pos_with_arg(end_position)
        if XSize is None or YSize is None or YSize is None:
            raise AttributeError("Dimension Argument must be a valid 3 Element List")
        if bond_type==0:
            bondwire = "JEDEC_5Points"
        elif bond_type==1:
            bondwire = "JEDEC_4Points"

        elif bond_type==2:
            bondwire = "LOW"
        else:
            self.messenger.add_error_message("Wrong Profile Type")
            return False
        vArg1 = ["NAME:BondwireParameters"]
        vArg1.append("WireType:="), vArg1.append(bondwire)
        vArg1.append("WireDiameter:="), vArg1.append(self.arg_with_dim(diameter))
        vArg1.append("NumSides:="), vArg1.append(str(facets))
        vArg1.append("XPadPos:="), vArg1.append(XPosition)
        vArg1.append("YPadPos:="), vArg1.append(YPosition)
        vArg1.append("ZPadPos:="), vArg1.append(ZPosition)
        vArg1.append("XDir:="), vArg1.append(XSize)
        vArg1.append("YDir:="), vArg1.append(YSize)
        vArg1.append("ZDir:="), vArg1.append(ZSize)
        vArg1.append("Distance:="), vArg1.append(
            self.arg_with_dim(GeometryOperators.points_distance(start_position, end_position)))
        vArg1.append("h1:="), vArg1.append(self.arg_with_dim(h1))
        vArg1.append("h2:="), vArg1.append(self.arg_with_dim(h2))
        vArg1.append("alpha:="), vArg1.append(self.arg_with_dim(alpha, "deg"))
        vArg1.append("beta:="), vArg1.append(self.arg_with_dim(beta, "deg"))
        vArg1.append("WhichAxis:="), vArg1.append("Z")
        vArg1.append("ReverseDirection:="), vArg1.append(False)
        vArg2 = o.export_attributes(name)
        o._m_name =self.oeditor.CreateBondwire(vArg1, vArg2)

        self._refresh_object_types()
        id = self._update_object(o)
        return id



    @aedt_exception_handler
    def create_region(self, pad_percent):
        if "Region" in self.get_all_objects_names():
            return None
        o = self._new_object()

        arg = ["NAME:RegionParameters"]
        p = ["+X", "+Y", "+Z", "-X", "-Y", "-Z"]
        i = 0
        for pval in p:
            pvalstr = str(pval) + "PaddingType:="
            qvalstr = str(pval) + "Padding:="
            arg.append(pvalstr)
            arg.append("Percentage Offset")
            arg.append(qvalstr)
            arg.append(str(pad_percent[i]))
            i += 1
        arg2 = ["NAME:Attributes", "Name:=", "Region", "Flags:=", "Wireframe#", "Color:=", "(143 175 143)",
                "Transparency:=", 0, "PartCoordinateSystem:=", "Global", "UDMId:=", "", "MaterialValue:=",
                "\"air\"", "SurfaceMaterialValue:=", "\"\"", "SolveInside:=", True, "IsMaterialEditable:=", True,
                "UseMaterialAppearance:=", False, "IsLightweight:=", False]

        self.oeditor.CreateRegion(arg, arg2)
        #TODO put this into Object3d Constructor?
        o._m_name = "Region"
        o._solve_inside = True
        o._transparency = 0
        o._wireframe = True

        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_circle(self, cs_plane, position, radius, numSides=0, name=None, matname=None):
        """Create a circle

        Parameters
        ----------
        cs_plane :
            ApplicationName.CoordinateSystemPlane object
        position :
            ApplicationName.modeler.Position(x,y,z) object
        radius :
            radius float
        numSides :
            Number of sides. 0 for circle (Default value = 0)
        name :
            Object Name (Default value = None)
        matname :
            material name. Optional, if nothing default material will be assigned

        Returns
        -------
        type
            id

        """
        o = self._new_object(matname=matname)

        szAxis = GeometryOperators.cs_plane_str(cs_plane)
        XCenter, YCenter, ZCenter = self.pos_with_arg(position)
        Radius = self.arg_with_dim(radius)


        vArg1 = ["NAME:CircleParameters"]
        vArg1.append("XCenter:="), vArg1.append(XCenter)
        vArg1.append("YCenter:="), vArg1.append(YCenter)
        vArg1.append("ZCenter:="), vArg1.append(ZCenter)
        vArg1.append("Radius:="), vArg1.append(Radius)
        vArg1.append("WhichAxis:="), vArg1.append(szAxis)
        vArg1.append("NumSegments:="), vArg1.append('{}'.format(numSides))

        vArg2 = o.export_attributes(name)

        o._m_name =self.oeditor.CreateCircle(vArg1, vArg2)

        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_sphere(self, position, radius, name=None, matname=None):
        """Create a sphere

        Parameters
        ----------
        position :
            ApplicationName.modeler.Position(x,y,z) object
        radius :
            radius float
        name :
            Object Name (Default value = None)
        matname :
            material name. Optional, if nothing default material will be assigned

        Returns
        -------

        """
        o = self._new_object(matname=matname)

        XCenter, YCenter, ZCenter = self.pos_with_arg(position)

        Radius = self.arg_with_dim(radius)

        vArg1 = ["NAME:SphereParameters"]
        vArg1.append("XCenter:="), vArg1.append(XCenter)
        vArg1.append("YCenter:="), vArg1.append(YCenter)
        vArg1.append("ZCenter:="), vArg1.append(ZCenter)
        vArg1.append("Radius:="), vArg1.append(Radius)

        vArg2 = o.export_attributes(name)

        o._m_name =self.oeditor.CreateSphere(vArg1, vArg2)

        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_cylinder(self, cs_axis, position, radius, height, numSides=0, name=None, matname=None):
        """Create a cylinder

        Parameters
        ----------
        cs_axis :
            ApplicationName.CoordinateSystemAxis object
        position :
            ApplicationName.modeler.Position(x,y,z) object
        radius :
            radius float
        height :
            height float
        numSides :
            Number of sides. 0 for circle (Default value = 0)
        name :
            Object Name (Default value = None)
        matname :
            material name. Optional, if nothing default material will be assigned

        Returns
        -------

        """
        o = self._new_object(matname=matname)

        szAxis = GeometryOperators.cs_axis_str(cs_axis)
        XCenter, YCenter, ZCenter = self.pos_with_arg(position)

        Radius = self.arg_with_dim(radius)
        Height = self.arg_with_dim(height)

        vArg1 = ["NAME:CylinderParameters"]
        vArg1.append("XCenter:="), vArg1.append(XCenter)
        vArg1.append("YCenter:="), vArg1.append(YCenter)
        vArg1.append("ZCenter:="), vArg1.append(ZCenter)
        vArg1.append("Radius:="), vArg1.append(Radius)
        vArg1.append("Height:="), vArg1.append(Height)
        vArg1.append("WhichAxis:="), vArg1.append(szAxis)
        vArg1.append("NumSides:="), vArg1.append('{}'.format(numSides))

        vArg2 = o.export_attributes(name)

        o._m_name =self.oeditor.CreateCylinder(vArg1, vArg2)

        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_ellipse(self, cs_plane, position, major_raidus, ratio, bIsCovered=True, name=None, matname=None):
        """Create a ellipse

        Parameters
        ----------
        cs_plane :
            ApplicationName.CoordinateSystemPlane object
        position :
            ApplicationName.modeler.Position(x,y,z) object
        major_raidus :
            radius float
        ratio :
            Ratio float
        bIsCovered :
            Boolean (Default value = True)
        name :
            Object Name (Default value = None)
        matname :
            material name. Optional, if nothing default material will be assigned

        Returns
        -------

        """
        o = self._new_object(matname=matname)

        szAxis = GeometryOperators.cs_plane_str(cs_plane)
        XStart, YStart, ZStart = self.pos_with_arg(position)

        MajorRadius = self.arg_with_dim(major_raidus)
        # Ratio = self.arg_with_dim(ratio)
        Ratio = ratio

        vArg1 = ["NAME:EllipseParameters"]
        vArg1.append("IsCovered:="), vArg1.append(bIsCovered)
        vArg1.append("XCenter:="), vArg1.append(XStart)
        vArg1.append("YCenter:="), vArg1.append(YStart)
        vArg1.append("ZCenter:="), vArg1.append(ZStart)
        vArg1.append("MajRadius:="), vArg1.append(MajorRadius)
        vArg1.append("Ratio:="), vArg1.append(Ratio)
        vArg1.append("WhichAxis:="), vArg1.append(szAxis)

        vArg2 = o.export_attributes(name)
        assert vArg2
        o._m_name =self.oeditor.CreateEllipse(vArg1, vArg2)

        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_equationbased_curve(self, udpequationbasedcurveddefinition, name=None, matname=None):
        """

        Parameters
        ----------
        udpequationbasedcurveddefinition :

        name :
             (Default value = None)
        matname :
             (Default value = None)

        Returns
        -------

        """
        o = self._new_object(matname=matname)

        vArg1 = udpequationbasedcurveddefinition.toScript()
        vArg2 = o.export_attributes(name)

        o._m_name =self.oeditor.CreateEquationCurve(vArg1, vArg2)

        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_helix(self, udphelixdefinition):
        """

        Parameters
        ----------
        udphelixdefinition :


        Returns
        -------

        """
        # TODO Test
        id = udphelixdefinition.ProfileID

        o = self.objects[id]

        vArg1 = ["NAME:Selections"]
        vArg1.append("Selections:="), vArg1.append(o.name)
        vArg1.append("NewPartsModelFlag:="), vArg1.append('Model')

        vArg2 = udphelixdefinition.toScript(self.model_units)

        self.oeditor.CreateHelix(vArg1, vArg2)
        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def convert_segments_to_line(self, object_name):
        """Convert a CreatePolyline list of segment into lines. it applies to splines and 3 points args

        Parameters
        ----------
        object_name :
            object id or object name

        Returns
        -------
        type
            Bool

        """
        if type(object_name) is int:
            object_name = self.objects[object_name].name
        edges = self.get_object_edges(object_name)
        for i in reversed(range(len(edges))):
            self.oeditor.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:Geometry3DPolylineTab",
                        [
                            "NAME:PropServers",
                            object_name + ":CreatePolyline:1:Segment" + str(i)
                        ],
                        [
                            "NAME:ChangedProps",
                            [
                                "NAME:Segment Type",
                                "Value:=", "Line"
                            ]
                        ]
                    ]
                ])
        return True

    @aedt_exception_handler
    def create_rectangle(self, csPlane, position, dblList, name=None, matname=None):
        """Create a rectangle

        Parameters
        ----------
        cs_plane : CoordinateSystemPlane
            CoordinateSystem Plane
        position :
            ApplicationName.modeler.Position(x,y,z) object
        dimension_list :
            dimension list
        name :
            Object Name (Default value = None)
        matname :
            material name. Optional, if nothing default material will be assigned

        Returns
        -------
        type
            id

        """
        o = self.request_new_object(matname=matname)

        szAxis = GeometryOperators.cs_plane_str(csPlane)
        XStart, YStart, ZStart = self.pos_with_arg(position)

        Width = self.arg_with_dim(dblList[0])
        Height = self.arg_with_dim(dblList[1])

        vArg1 = ["NAME:RectangleParameters"]
        vArg1.append("XStart:="), vArg1.append(XStart)
        vArg1.append("YStart:="), vArg1.append(YStart)
        vArg1.append("ZStart:="), vArg1.append(ZStart)
        vArg1.append("Width:="), vArg1.append(Width)
        vArg1.append("Height:="), vArg1.append(Height)
        vArg1.append("WhichAxis:="), vArg1.append(szAxis)

        vArg2 = o.export_attributes(name)
        o._m_name =self.oeditor.CreateRectangle(vArg1, vArg2)
        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def create_udm(self, udmfullname, udm_params_list, udm_library='syslib'):
        """Create User Defined Model

        Parameters
        ----------
        udmfullname :
            udm full name including folder
        udm_params_list :
            udpm pairs object
        udm_library :
            udp library (Default value = 'syslib')

        Returns
        -------
        type
            object ID

        """


        vArg1 = ["NAME:UserDefinedModelParameters",["NAME:Definition"], ["NAME:Options"]]
        vArgParamVector = ["NAME:GeometryParams"]

        for pair in udm_params_list:
            if type(pair) is list:
                name = pair[0]
                val = pair[1]
            else:
                name = pair.Name
                val = pair.Value
            if type(val) is int:
                vArgParamVector.append(
                    ["NAME:UDMParam", "Name:=", name, "Value:=", str(val), "PropType2:=", 3, "PropFlag2:=", 2])
            elif str(val)[0] in '0123456789':
                vArgParamVector.append(
                    ["NAME:UDMParam", "Name:=", name, "Value:=", str(val), "PropType2:=", 3, "PropFlag2:=", 4])
            else:
                vArgParamVector.append(
                    ["NAME:UDMParam", "Name:=", name, "Value:=", str(val), "DataType:=", "String", "PropType2:=", 1,
                     "PropFlag2:=", 0])

        vArg1.append(vArgParamVector)
        vArg1.append("DllName:=")
        vArg1.append(udmfullname)
        vArg1.append("Library:=")
        vArg1.append(udm_library)
        vArg1.append("Version:=")
        vArg1.append("2.0")
        vArg1.append("ConnectionID:=")
        vArg1.append("")
        oname = self.oeditor.CreateUserDefinedModel(vArg1)
        if oname:
            object_lists = self.oeditor.GetPartsForUserDefinedModel(oname)
            for el in object_lists:
                o = self._new_object()
                o._m_name =el
                if el in list(self.oeditor.GetObjectsInGroup("Solids")):
                    id = self._update_object(o)
                elif el in list(self.oeditor.GetObjectsInGroup("Sheets")):
                    id = self._update_object(o)
                elif el in list(self.oeditor.GetObjectsInGroup("Lines")):
                    id = self._update_object(o)
                else:
                    id = self._update_object(o)
                self.update_object_properties(o)
            return oname
        else:
            return False

    @aedt_exception_handler
    def create_cone(self, cs_axis, position, bottom_radius, top_radius, height, name=None, matname=None):
        """Create a cone

        Parameters
        ----------
        cs_axis : CoordinateSystemAxis
            CoordinateSystem Axis
        position :
            ApplicationName.modeler.Position(x,y,z) object
        bottom_radius :
            bottom radius
        top_radius :
            topradius radius
        height :
            height
        name :
            Object Name (Default value = None)
        matname :
            material name. Optional, if nothing default material will be assigned

        Returns
        -------
        type
            id

        """
        o = self._new_object(matname=matname)

        XCenter, YCenter, ZCenter = self.pos_with_arg(position)
        szAxis = GeometryOperators.cs_axis_str(cs_axis)
        Height = self.arg_with_dim(height)
        RadiusBt = self.arg_with_dim(bottom_radius)
        RadiusUp = self.arg_with_dim(top_radius)

        vArg1 = ["NAME:ConeParameters"]
        vArg1.append("XCenter:="), vArg1.append(XCenter)
        vArg1.append("YCenter:="), vArg1.append(YCenter)
        vArg1.append("ZCenter:="), vArg1.append(ZCenter)
        vArg1.append("WhichAxis:="), vArg1.append(szAxis)
        vArg1.append("Height:="), vArg1.append(Height)
        vArg1.append("BottomRadius:="), vArg1.append(RadiusBt)
        vArg1.append("TopRadius:="), vArg1.append(RadiusUp)

        vArg2 = o.export_attributes(name)

        o._m_name =self.oeditor.CreateCone(vArg1, vArg2)

        if o.analysis_type:
            o.material_name, o.solve_inside = self._check_material(matname, self.defaultmaterial)

        self._refresh_object_types()
        id = self._update_object(o)
        return id

    @aedt_exception_handler
    def insert_3d_component(self, compFile, geoParams, szMatParams='', szDesignParams='', targetCS='Global'):
        """Insert a new 3d Component object

        Parameters
        ----------
        compFile :
            Component file
        geoParams :
            Geometrical Parameters
        szMatParams :
            Material Parameters (Default value = '')
        szDesignParams :
            Design Parameters (Default value = '')
        targetCS :
            Target Coordinate System (Default value = 'Global')

        Returns
        -------

        """
        o = self._new_object()

        vArg1 = ["NAME:InsertComponentData"]

        szGeoParams = ''
        # for pair in geoParams:
        #     name = pair.Name
        #     val = pair.Value
        #     szGeoParams += "{0}='{1}' ".format(name, val)
        for par in geoParams:
            name = par
            val = geoParams[par]
            szGeoParams += "{0}='{1}' ".format(name, val)

        vArg1.append("GeometryParameters:=")
        vArg1.append(szGeoParams)
        vArg1.append("MaterialParameters:=")
        vArg1.append(szMatParams)
        vArg1.append("DesignParameters:=")
        vArg1.append(szDesignParams)
        vArg1.append("TargetCS:=")
        vArg1.append(targetCS)
        vArg1.append("ComponentFile:=")
        vArg1.append(compFile)

        if self.oeditor is not None:
            o._m_name =self.oeditor.Insert3DComponent(vArg1)
            self.refresh_all_ids()
        return id

    @aedt_exception_handler
    def get_3d_component_object_list(self, componentname):
        """Given a 3DComponent it returns the list of all objects belonging to it

        Parameters
        ----------
        componentname :
            3DComponent Name

        Returns
        -------
        type
            List of objects

        """
        compobj = self.oeditor.GetChildObject(componentname)
        if compobj:
            return list(compobj.GetChildNames())
        else:
            return []

    @aedt_exception_handler
    def get_all_solids_names(self, refresh_list=False):
        """get all solids names in the design

        Parameters
        ----------
        refresh_list :
            bool, force the refresh of objects (Default value = False)

        Returns
        -------
        type
            list of the solids names

        """
        return self.get_all_objects_names(refresh_list=refresh_list, get_solids=True, get_sheets=False, get_lines=False)

    @aedt_exception_handler
    def get_face_normal(self, faceId, bounding_box=None):
        """Get the face normal.
        Limitations:
        - the face must be planar.
        - Currently it works only if the face has at least two vertices. Notable excluded items are circles and
        ellipses that have only one vertex.
        - If a bounding box is specified, the normal is orientated outwards with respect to the bounding box.
        Usually the bounding box refers to a volume where the face lies.
        If no bounding box is specified, the normal can be inward or outward the volume.

        Parameters
        ----------
        faceId :
            part ID (integer).
        bounding_box :
            optional, bounding box in the form [x_min, y_min, z_min, x_Max, y_Max, z_Max] (Default value = None)

        Returns
        -------
        type
            normal versor (normalized [x, y, z]) or None.

        """
        vertices_ids = self.get_face_vertices(faceId)
        if len(vertices_ids) < 2:
            return None
        else:
            v1 = self.get_vertex_position(vertices_ids[0])
            v2 = self.get_vertex_position(vertices_ids[1])
        fc = self.get_face_center(faceId)
        cv1 = GeometryOperators.v_points(fc, v1)
        cv2 = GeometryOperators.v_points(fc, v2)
        n = GeometryOperators.v_cross(cv1, cv2)
        normal = GeometryOperators.normalize_vector(n)
        if not bounding_box:
            return normal
        else:
            """
            try to move the face center twice, the first with the normal vector, and the second with its inverse.
            Measures which is closer to the center point of the bounding box.
            """
            inv_norm = [-i for i in normal]
            mv1 = GeometryOperators.v_sum(fc, normal)
            mv2 = GeometryOperators.v_sum(fc, inv_norm)
            bb_center = GeometryOperators.get_mid_point(bounding_box[0:3], bounding_box[3:6])
            d1 = GeometryOperators.points_distance(mv1, bb_center)
            d2 = GeometryOperators.points_distance(mv2, bb_center)
            if d1 > d2:
                return normal
            else:
                return inv_norm
